# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
It defines subcommands for verdi group command.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import click

from aiida.common.exceptions import UniquenessError
from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.utils import echo
from aiida.cmdline.utils.decorators import with_dbenv
from aiida.cmdline.params import options, arguments


@verdi.group('group')
def verdi_group():
    """Inspect, create and manage groups."""
    pass


@verdi_group.command("removenodes")
@options.GROUP()
@arguments.NODES()
@options.FORCE(help="Force to remove the nodes from group.")
@with_dbenv()
def group_removenodes(group, nodes, force):
    """
    Remove NODES from a given AiiDA group.
    """
    if not force:
        click.confirm(
            "Are you sure to remove {} nodes from the group with PK = {} "
            "({})?".format(len(nodes), group.pk, group.name),
            abort=True)

    group.remove_nodes(nodes)


@verdi_group.command("addnodes")
@options.GROUP()
@options.FORCE(help="Force to add nodes in the group.")
@arguments.NODES()
@with_dbenv()
def group_addnodes(group, force, nodes):
    """
    Add NODES to a given AiiDA group.
    """

    if not force:
        click.confirm(
            "Are you sure to add {} nodes the group with PK = {} "
            "({})?".format(len(nodes), group.pk, group.name),
            abort=True)

    group.add_nodes(nodes)


@verdi_group.command("delete")
@arguments.GROUP()
@options.FORCE(help="Force deletion of the group even if it "
               "is not empty. Note that this deletes only the "
               "group and not the nodes.")
@with_dbenv()
def group_delete(group, force):
    """
    Pass the GROUP to delete an existing group.
    """
    group_pk = group.pk
    group_name = group.name

    num_nodes = len(group.nodes)
    if num_nodes > 0 and not force:
        echo.echo_critical(("Group '{}' is not empty (it contains {} "
                            "nodes). Pass the -f option if you really want to delete "
                            "it.".format(group_name, num_nodes)))

    if not force:
        click.confirm('Are you sure to kill the group with PK = {} ({})?'.format(group_pk, group_name), abort=True)

    group.delete()
    echo.echo_success("Group '{}' (PK={}) deleted.".format(group_name, group_pk))


@verdi_group.command("rename")
@arguments.GROUP()
@click.argument("name", nargs=1, type=click.STRING)
@with_dbenv()
def group_rename(group, name):
    """
    Rename an existing group. Pass the GROUP for which you want to rename and its
    new NAME.
    """
    try:
        group.name = name
    except UniquenessError as exception:
        echo.echo_critical("Error: {}.".format(exception))
    else:
        echo.echo_success('Name successfully changed')


@verdi_group.command("description")
@arguments.GROUP()
@click.argument("description", type=click.STRING)
@with_dbenv()
def group_description(group, description):
    """
    Change the description of a given group.
    Pass the GROUP for which you want to edit the description and its
    new DESCRIPTION. If DESCRIPTION is not provided, just show the current description.

    """
    group.description = description


@verdi_group.command("show")
@options.RAW(help="Show only a space-separated list of PKs of the calculations in the group")
@click.option(
    '-u',
    '--uuid',
    is_flag=True,
    default=False,
    help="Show UUIDs together with PKs. Note: if the --raw option is also passed, PKs are not printed, but oly UUIDs.")
@arguments.GROUP()
@with_dbenv()
def group_show(group, raw, uuid):
    """
    Show information on a given group. Pass the GROUP as a parameter.
    """

    from aiida.common.utils import str_timedelta
    from aiida.utils import timezone
    from aiida.plugins.loader import get_plugin_type_from_type_string
    from tabulate import tabulate

    if raw:
        if uuid:
            echo.echo(" ".join(str(_.uuid) for _ in group.nodes))
        else:
            echo.echo(" ".join(str(_.pk) for _ in group.nodes))
    else:
        type_string = group.type_string
        desc = group.description
        now = timezone.now()

        table = []
        table.append(["Group name", group.name])
        table.append(["Group type", type_string if type_string else "<user-defined>"])
        table.append(["Group description", desc if desc else "<no description>"])
        echo.echo(tabulate(table))

        table = []
        header = []
        if uuid:
            header.append('UUID')
        header.extend(['PK', 'Type', 'Created'])
        echo.echo("# Nodes:")
        for node in group.nodes:
            row = []
            if uuid:
                row.append(node.uuid)
            row.append(node.pk)
            row.append(get_plugin_type_from_type_string(node.type).rsplit(".", 1)[1])
            row.append(str_timedelta(now - node.ctime, short=True, negative_to_zero=True))
            table.append(row)
        echo.echo(tabulate(table, headers=header))


# pylint: disable=too-many-arguments, too-many-locals
@verdi_group.command("list")
@options.ALL_USERS(help="Show groups for all users, rather than only for the current user")
@click.option(
    '-u',
    '--user',
    'user_email',
    type=click.STRING,
    help="Add a filter to show only groups belonging to a specific user")
@click.option(
    '-t',
    '--type',
    'group_type',
    type=click.STRING,
    help="Show groups of a specific type, instead of user-defined groups")
@click.option(
    '-d', '--with-description', 'with_description', is_flag=True, default=False, help="Show also the group description")
@click.option('-C', '--count', is_flag=True, default=False, help="Show also the number of nodes in the group")
@options.PAST_DAYS(help="add a filter to show only groups created in the past N days", default=None)
@click.option(
    '-s',
    '--startswith',
    type=click.STRING,
    default=None,
    help="add a filter to show only groups for which the name begins with STRING")
@click.option(
    '-e',
    '--endswith',
    type=click.STRING,
    default=None,
    help="add a filter to show only groups for which the name ends with STRING")
@click.option(
    '-c',
    '--contains',
    type=click.STRING,
    default=None,
    help="add a filter to show only groups for which the name contains STRING")
@options.NODE(help="Show only the groups that contain the node")
@with_dbenv()
def group_list(all_users, user_email, group_type, with_description, count, past_days, startswith, endswith, contains,
               node):
    """
    List AiiDA user-defined groups.
    """
    import datetime
    from aiida.utils import timezone
    from aiida.orm.group import get_group_type_mapping
    from aiida.orm.backend import construct_backend
    from tabulate import tabulate

    if all_users and user_email is not None:
        echo.echo_critical("argument -A/--all-users: not allowed with argument -u/--user")

    backend = construct_backend()

    if all_users:
        user = None
    else:
        if user_email:
            user = user_email
        else:
            # By default: only groups of this user
            user = backend.users.get_automatic_user()

    type_string = ""
    if group_type is not None:
        try:
            type_string = get_group_type_mapping()[group_type]
        except KeyError:
            echo.echo_critical("Invalid group type. Valid group types are: {}".format(",".join(
                sorted(get_group_type_mapping().keys()))))

    n_days_ago = None
    if past_days:
        n_days_ago = (timezone.now() - datetime.timedelta(days=past_days))

    name_filters = {'startswith': startswith, 'endswith': endswith, 'contains': contains}

    # Depending on --nodes option use or not key "nodes"
    from aiida.orm.implementation import Group as OrmGroup

    if node:
        result = OrmGroup.query(
            user=user, type_string=type_string, nodes=node, past_days=n_days_ago, name_filters=name_filters)
    else:
        result = OrmGroup.query(user=user, type_string=type_string, past_days=n_days_ago, name_filters=name_filters)

    projection_lambdas = {
        'pk': lambda group: str(group.pk),
        'name': lambda group: group.name,
        'count': lambda group: len(group.nodes),
        'user': lambda group: group.user.email.strip(),
        'description': lambda group: group.description
    }

    table = []
    projection_header = ['PK', 'Name', 'User']
    projection_fields = ['pk', 'name', 'user']

    if with_description:
        projection_header.append('Description')
        projection_fields.append('description')

    if count:
        projection_header.append('Node count')
        projection_fields.append('count')

    for group in result:
        table.append([projection_lambdas[field](group) for field in projection_fields])

    echo.echo(tabulate(table, headers=projection_header))


@verdi_group.command("create")
@click.argument('group_name', nargs=1, type=click.STRING)
@with_dbenv()
def group_create(group_name):
    """
    Create a new empty group with the name GROUP_NAME
    """

    from aiida.orm import Group as OrmGroup

    group, created = OrmGroup.get_or_create(name=group_name)

    if created:
        echo.echo_success("Group created with PK = {} and name '{}'".format(group.pk, group.name))
    else:
        echo.echo_info("Group '{}' already exists, PK = {}".format(group.name, group.pk))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Module containing utilities and classes relating to job calculations running
on systems that require transport.
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import contextlib
from functools import partial
import time
from six import iteritems, itervalues
from tornado import concurrent, gen

from aiida import scheduler as schedulers
from aiida.common import exceptions
from . import utils


class JobsList(object):
    """
    A list of submitted jobs on a machine connected to by transport based on the
    authorisation information.
    """

    def __init__(self, authinfo, transport_queue):
        """
        :param authinfo: The authinfo used to check the jobs list
        :type authinfo: :class:`aiida.orm.AuthInfo`
        :param transport_queue: A transport queue
        :type: :class:`aiida.work.transports.TransportQueue`
        """
        self._authinfo = authinfo
        self._transport_queue = transport_queue
        self._loop = transport_queue.loop()

        self._jobs_cache = {}
        self._last_updated = None  # type: float
        self._job_update_requests = {}  # Mapping: {job_id: Future}
        self._update_handle = None

    def get_minimum_update_interval(self):
        """
        Get the minimum interval that can be expected between updates of the list
        :return: The minimum interval
        :rtype: float
        """
        return self._authinfo.computer.get_minimum_job_poll_interval()

    def get_last_updated(self):
        """
        Get the timestamp of when the list was last updated as produced by `time.time()`

        :return: The last update point
        :rtype: float
        """
        return self._last_updated

    @gen.coroutine
    def _get_jobs_from_scheduler(self):
        """
        Get the current jobs list from the scheduler

        :return: A dictionary of {job_id: job info}
        :rtype: dict
        """
        with self._transport_queue.request_transport(self._authinfo) as request:
            transport = yield request

            scheduler = self._authinfo.computer.get_scheduler()
            scheduler.set_transport(transport)

            kwargs = {'as_dict': True}
            if scheduler.get_feature('can_query_by_user'):
                kwargs['user'] = "$USER"
            else:
                kwargs['jobs'] = self._get_jobs_with_scheduler()

            scheduler_response = scheduler.getJobs(**kwargs)
            jobs_cache = {}

            for job_id, job_info in iteritems(scheduler_response):
                # If the job is done then get detailed job information
                detailed_job_info = None
                if job_info.job_state == schedulers.JOB_STATES.DONE:
                    try:
                        detailed_job_info = scheduler.get_detailed_jobinfo(job_id)
                    except exceptions.FeatureNotAvailable:
                        detailed_job_info = 'This scheduler does not implement get_detailed_jobinfo'

                job_info.detailedJobinfo = detailed_job_info
                jobs_cache[job_id] = job_info

            raise gen.Return(jobs_cache)

    @gen.coroutine
    def _update_job_info(self):
        """
        Update all of the job information objects for a given authinfo, that is to say for
        all the jobs on a particular machine for a particular user.

        This will set the futures for all pending update requests where the corresponding job
        has a new status compared to the last update.
        """
        try:
            if not self._update_requests_outstanding():
                return

            # Update our cache of the job states
            self._jobs_cache = yield self._get_jobs_from_scheduler()
        except Exception as exception:
            # Set the exception on all the update futures
            for future in itervalues(self._job_update_requests):
                if not future.done():
                    future.set_exception(exception)
            raise
        else:
            for job_id, future in iteritems(self._job_update_requests):
                if not future.done():
                    future.set_result(self._jobs_cache.get(job_id, None))
        finally:
            self._job_update_requests = {}

    @contextlib.contextmanager
    def request_job_info_update(self, job_id):
        """
        Request job info about a job when it next changes it's job state.  If the job is not
        found in the jobs list at the update the future will resolve to None.

        :param job_id: The job identifier
        :return: A future that will resolve to a JobInfo object when the job changes state
        """
        # Get or create the future
        request = self._job_update_requests.setdefault(job_id, concurrent.Future())
        assert not request.done(), "The future should be no be in the done state"

        try:
            self._ensure_updating()
            yield request
        finally:
            pass

    def _ensure_updating(self):
        """
        Ensure that we are updating the job list from the remote resource.
        This will automatically stop if there are no outstanding requests.
        """

        @gen.coroutine
        def updating():
            """ Do the actual update, stop if not requests left """
            yield self._update_job_info()
            # Any outstanding requests?
            if self._update_requests_outstanding():
                self._update_handle = self._loop.call_later(self._get_next_update_delay(), updating)
            else:
                self._update_handle = None

        # Check if we're already updating
        if self._update_handle is None:
            self._update_handle = self._loop.call_later(self._get_next_update_delay(), updating)

    @staticmethod
    def _has_job_state_changed(old, new):
        """
        :type old: :class:`aiida.scheduler.JobInfo`
        :type new: :class:`aiida.scheduler.JobInfo`
        :rtype: bool
        """
        if old is None and new is None:
            return False
        elif old is None or new is None:
            # One is None and the other isn't
            return True

        return old.job_state != new.job_state or \
               old.job_substate != new.job_substate

    def _get_next_update_delay(self):
        """
        Calculate when we are next allowed to call the scheduler get jobs command
        based on when we last called it, how long has elapsed and the minimum given
        update interval.

        :return: The delay (in seconds) for when it's safe to call the get jobs command
        :rtype: float
        """
        if self._last_updated is None:
            # Never updated, so do it straight away
            return 0.

        # Make sure to actually 'get' it here, so that if the user changed it
        # between times we use the current value
        minimum_interval = self._authinfo.computer.get_minimum_job_poll_interval()
        elapsed = time.time() - self._last_updated

        return max(minimum_interval - elapsed, 0.)

    def _update_requests_outstanding(self):
        return any(not request.done() for request in itervalues(self._job_update_requests))

    def _get_jobs_with_scheduler(self):
        """
        Get all the jobs that are currently with scheduler for this authinfo

        :return: the list of jobs with the scheduler
        :rtype: list
        """
        return [str(job_id) for job_id, _ in self._job_update_requests.items()]


class JobManager(object):
    """
    A manager for jobs on a (usually) remote resource such as a supercomputer
    """

    def __init__(self, transport_queue):
        self._transport_queue = transport_queue
        self._job_lists = utils.RefObjectStore()

    @contextlib.contextmanager
    def request_job_info_update(self, authinfo, job_id):
        """
        Get a future that will resolve to information about a given job.  This is a context
        manager so that if the user leaves the context the request is automatically cancelled.

        :return: A tuple containing the JobInfo object and detailed job info.  Both can be None.
        :rtype: :class:`tornado.concurrent.Future`
        """
        # Define a way to create a JobsList if needed
        create = partial(JobsList, authinfo, self._transport_queue)

        with self._job_lists.get(authinfo.id, create) as job_list:
            with job_list.request_job_info_update(job_id) as request:
                try:
                    yield request
                finally:
                    if not request.done():
                        request.cancel()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=protected-access,too-many-locals,invalid-name,too-many-public-methods
"""Tests for `verdi calculation`."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from click.testing import CliRunner

from aiida.backends.testbase import AiidaTestCase
from aiida.cmdline.commands import cmd_calculation as command
from aiida.common.datastructures import calc_states
from aiida.orm.calculation.job import JobCalculationExitStatus
from aiida.work import runners, rmq


def get_result_lines(result):
    return [e for e in result.output.split('\n') if e]


class TestVerdiCalculation(AiidaTestCase):
    """Tests for `verdi calculation`."""

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestVerdiCalculation, cls).setUpClass(*args, **kwargs)
        from aiida.backends.tests.utils.fixtures import import_archive_fixture
        from aiida.common.exceptions import ModificationNotAllowed
        from aiida.common.links import LinkType
        from aiida.orm import Code, Group, Node, JobCalculation, CalculationFactory
        from aiida.orm.data.parameter import ParameterData
        from aiida.orm.querybuilder import QueryBuilder
        from aiida.work.processes import ProcessState
        from aiida.orm.backend import construct_backend
        backend = construct_backend()

        rmq_config = rmq.get_rmq_config()

        # These two need to share a common event loop otherwise the first will never send
        # the message while the daemon is running listening to intercept
        cls.runner = runners.Runner(rmq_config=rmq_config, rmq_submit=True, poll_interval=0.)
        cls.daemon_runner = runners.DaemonRunner(rmq_config=rmq_config, rmq_submit=True, poll_interval=0.)

        cls.computer = backend.computers.create(name='comp', hostname='localhost', transport_type='local',
                                                scheduler_type='direct', workdir='/tmp/aiida').store()

        cls.code = Code(remote_computer_exec=(cls.computer, '/bin/true')).store()
        cls.group = Group(name='test_group').store()
        cls.node = Node().store()
        cls.calcs = []

        authinfo = backend.authinfos.create(computer=cls.computer, user=backend.users.get_automatic_user())
        authinfo.store()

        # Create 13 JobCalculations (one for each CalculationState)
        for calculation_state in calc_states:

            calc = JobCalculation(
                computer=cls.computer, resources={
                    'num_machines': 1,
                    'num_mpiprocs_per_machine': 1
                }).store()

            # Trying to set NEW will raise, but in this case we don't need to change the state
            try:
                calc._set_state(calculation_state)
            except ModificationNotAllowed:
                pass

            try:
                exit_status = JobCalculationExitStatus[calculation_state]
            except KeyError:
                if calculation_state == 'IMPORTED':
                    calc._set_process_state(ProcessState.FINISHED)
                else:
                    calc._set_process_state(ProcessState.RUNNING)
            else:
                calc._set_exit_status(exit_status)
                calc._set_process_state(ProcessState.FINISHED)

            cls.calcs.append(calc)

            if calculation_state == 'PARSING':
                cls.KEY_ONE = 'key_one'
                cls.KEY_TWO = 'key_two'
                cls.VAL_ONE = 'val_one'
                cls.VAL_TWO = 'val_two'

                output_parameters = ParameterData(dict={
                    cls.KEY_ONE: cls.VAL_ONE,
                    cls.KEY_TWO: cls.VAL_TWO,
                }).store()

                output_parameters.add_link_from(calc, 'output_parameters', link_type=LinkType.RETURN)

                # Create shortcut for easy dereferencing
                cls.result_job = calc

                # Add a single calc to a group
                cls.group.add_nodes([calc])

        # Load the fixture containing a single ArithmeticAddCalculation node
        import_archive_fixture('calculation/simpleplugins.arithmetic.add.aiida')

        # Get the imported ArithmeticAddCalculation node
        ArithmeticAddCalculation = CalculationFactory('simpleplugins.arithmetic.add')
        calculations = QueryBuilder().append(ArithmeticAddCalculation).all()[0]
        cls.arithmetic_job = calculations[0]

    def setUp(self):
        self.cli_runner = CliRunner()

    def test_calculation_res(self):
        """Test verdi calculation res"""
        options = [str(self.result_job.uuid)]
        result = self.cli_runner.invoke(command.calculation_res, options)
        self.assertIsNone(result.exception)
        self.assertIn(self.KEY_ONE, result.output)
        self.assertIn(self.VAL_ONE, result.output)
        self.assertIn(self.KEY_TWO, result.output)
        self.assertIn(self.VAL_TWO, result.output)

        for flag in ['-k', '--keys']:
            options = [flag, self.KEY_ONE, '--', str(self.result_job.uuid)]
            result = self.cli_runner.invoke(command.calculation_res, options)
            self.assertIsNone(result.exception)
            self.assertIn(self.KEY_ONE, result.output)
            self.assertIn(self.VAL_ONE, result.output)
            self.assertNotIn(self.KEY_TWO, result.output)
            self.assertNotIn(self.VAL_TWO, result.output)

    def test_calculation_list_identifiers(self):
        """Test verdi calculation list with specific identifiers"""
        options = ['-r', '--project', 'pk', '--'] + [str(calc.pk) for calc in self.calcs[:2]]
        result = self.cli_runner.invoke(command.calculation_list, options)
        self.assertIsNone(result.exception)

        valid_pks = [calc.pk for calc in self.calcs]
        for line in get_result_lines(result):
            self.assertIn(int(line), valid_pks)

    def test_calculation_list_all_user(self):
        """Test verdi calculation list with the all users option"""
        for flag in ['-A', '--all-users']:
            options = ['-r', '-a', flag]
            result = self.cli_runner.invoke(command.calculation_list, options)
            self.assertIsNone(result.exception)
            self.assertEquals(len(get_result_lines(result)), 14)

    def test_calculation_list_all(self):
        """Test verdi calculation list with the all option"""

        # Without the flag I should only get the "active" states, which should be seven
        options = ['-r']
        result = self.cli_runner.invoke(command.calculation_list, options)
        self.assertIsNone(result.exception)
        self.assertEquals(len(get_result_lines(result)), 7, result.output)

        for flag in ['-a', '--all']:
            options = ['-r', flag]
            result = self.cli_runner.invoke(command.calculation_list, options)
            self.assertIsNone(result.exception)
            self.assertEquals(len(get_result_lines(result)), 13)

    def test_calculation_list_limit(self):
        """Test verdi calculation list with the limit option"""
        for flag in ['-l', '--limit']:
            limit = 1
            options = ['-r', flag, limit]
            result = self.cli_runner.invoke(command.calculation_list, options)
            self.assertIsNone(result.exception)
            self.assertEquals(len(get_result_lines(result)), limit)

    def test_calculation_list_group(self):
        """Test verdi calculation list with the group option"""
        for flag in ['-G', '--groups']:
            options = ['-r', '-a', flag, self.group.uuid]
            result = self.cli_runner.invoke(command.calculation_list, options)
            self.assertIsNone(result.exception)
            self.assertEquals(len(get_result_lines(result)), 1)

    def test_calculation_list_project(self):
        """Test verdi calculation list with the project option"""
        for flag in ['-P', '--project']:
            options = ['-r', flag, 'pk']
            result = self.cli_runner.invoke(command.calculation_list, options)
            self.assertIsNone(result.exception)

            valid_pks = [calc.pk for calc in self.calcs]
            for line in get_result_lines(result):
                self.assertIn(int(line), valid_pks)

    def test_calculation_list_calculation_state(self):
        """Test verdi calculation list with the calculation state filter"""
        for flag in ['-s', '--calculation-state']:
            for state in calc_states:
                # Each valid state should have exactly one entry
                options = ['-r', flag, state]
                result = self.cli_runner.invoke(command.calculation_list, options)
                self.assertIsNone(result.exception)
                self.assertEquals(len(get_result_lines(result)), 1)

    def test_calculation_list_process_state(self):
        """Test verdi calculation list with the process state filter"""
        for flag in ['-S', '--process-state']:
            for state in ['finished', 'running']:

                options = ['-r', flag, state]
                result = self.cli_runner.invoke(command.calculation_list, options)

                self.assertIsNone(result.exception)

                if state == 'finished':
                    self.assertEquals(len(get_result_lines(result)), 6, result.output)
                else:
                    self.assertEquals(len(get_result_lines(result)), 7, result.output)

    def test_calculation_list_failed(self):
        """Test verdi calculation list with the failed filter"""
        for flag in ['-X', '--failed']:
            options = ['-r', flag]
            result = self.cli_runner.invoke(command.calculation_list, options)

            self.assertIsNone(result.exception)
            self.assertEquals(len(get_result_lines(result)), 4, result.output)

    def test_calculation_list_exit_status(self):
        """Test verdi calculation list with the exit status filter"""
        for flag in ['-E', '--exit-status']:
            for exit_status in JobCalculationExitStatus:
                options = ['-r', flag, exit_status.value]
                result = self.cli_runner.invoke(command.calculation_list, options)

                self.assertIsNone(result.exception)
                self.assertEquals(len(get_result_lines(result)), 1)

    def test_calculation_show(self):
        """Test verdi calculation show"""

        # Running without identifiers should not except and not print anything
        options = []
        result = self.cli_runner.invoke(command.calculation_show, options)
        self.assertIsNone(result.exception)
        self.assertEquals(len(get_result_lines(result)), 0)

        # Giving a single identifier should print a non empty string message
        options = [str(self.calcs[0].pk)]
        result = self.cli_runner.invoke(command.calculation_show, options)
        self.assertIsNone(result.exception)
        self.assertTrue(len(get_result_lines(result)) > 0)

        # Giving multiple identifiers should print a non empty string message
        options = [str(calc.pk) for calc in self.calcs]
        result = self.cli_runner.invoke(command.calculation_show, options)
        self.assertIsNone(result.exception)
        self.assertTrue(len(get_result_lines(result)) > 0)

    def test_calculation_logshow(self):
        """Test verdi calculation logshow"""

        # Running without identifiers should not except and not print anything
        options = []
        result = self.cli_runner.invoke(command.calculation_logshow, options)
        self.assertIsNone(result.exception)
        self.assertEquals(len(get_result_lines(result)), 0)

        # Giving a single identifier should print a non empty string message
        options = [str(self.calcs[0].pk)]
        result = self.cli_runner.invoke(command.calculation_logshow, options)
        self.assertIsNone(result.exception)
        self.assertTrue(len(get_result_lines(result)) > 0)

        # Giving multiple identifiers should print a non empty string message
        options = [str(calc.pk) for calc in self.calcs]
        result = self.cli_runner.invoke(command.calculation_logshow, options)
        self.assertIsNone(result.exception)
        self.assertTrue(len(get_result_lines(result)) > 0)

    def test_calculation_plugins(self):
        """Test verdi calculation plugins"""
        from aiida.plugins.entry_point import get_entry_points

        calculation_plugins = get_entry_points('aiida.calculations')

        result = self.cli_runner.invoke(command.calculation_plugins, ['non_existent'])
        self.assertIsNotNone(result.exception)

        result = self.cli_runner.invoke(command.calculation_plugins, ['simpleplugins.arithmetic.add'])
        self.assertIsNone(result.exception)
        self.assertTrue(len(get_result_lines(result)) > 0)

        result = self.cli_runner.invoke(command.calculation_plugins)
        self.assertIsNone(result.exception)
        self.assertTrue(len(get_result_lines(result)) > len(calculation_plugins))

    def test_calculation_inputls(self):
        """Test verdi calculation inputls"""
        options = []
        result = self.cli_runner.invoke(command.calculation_inputls, options)
        self.assertIsNotNone(result.exception)

        options = [self.arithmetic_job.uuid]
        result = self.cli_runner.invoke(command.calculation_inputls, options)
        self.assertIsNone(result.exception)
        self.assertEquals(len(get_result_lines(result)), 3)
        self.assertIn('.aiida', get_result_lines(result))
        self.assertIn('aiida.in', get_result_lines(result))
        self.assertIn('_aiidasubmit.sh', get_result_lines(result))

        options = [self.arithmetic_job.uuid, '.aiida']
        result = self.cli_runner.invoke(command.calculation_inputls, options)
        self.assertIsNone(result.exception)
        self.assertEquals(len(get_result_lines(result)), 2)
        self.assertIn('calcinfo.json', get_result_lines(result))
        self.assertIn('job_tmpl.json', get_result_lines(result))

    def test_calculation_outputls(self):
        """Test verdi calculation outputls"""
        options = []
        result = self.cli_runner.invoke(command.calculation_outputls, options)
        self.assertIsNotNone(result.exception)

        options = [self.arithmetic_job.uuid]
        result = self.cli_runner.invoke(command.calculation_outputls, options)
        self.assertIsNone(result.exception)
        self.assertEquals(len(get_result_lines(result)), 3)
        self.assertIn('_scheduler-stderr.txt', get_result_lines(result))
        self.assertIn('_scheduler-stdout.txt', get_result_lines(result))
        self.assertIn('aiida.out', get_result_lines(result))

        options = [self.arithmetic_job.uuid, 'aiida.out']
        result = self.cli_runner.invoke(command.calculation_outputls, options)
        self.assertIsNone(result.exception)
        self.assertEquals(len(get_result_lines(result)), 1)
        self.assertIn('aiida.out', get_result_lines(result))

    def test_calculation_inputcat(self):
        """Test verdi calculation inputcat"""
        options = []
        result = self.cli_runner.invoke(command.calculation_inputcat, options)
        self.assertIsNotNone(result.exception)

        options = [self.arithmetic_job.uuid]
        result = self.cli_runner.invoke(command.calculation_inputcat, options)
        self.assertIsNone(result.exception)
        self.assertEquals(len(get_result_lines(result)), 1)
        self.assertEquals(get_result_lines(result)[0], '2 3')

        options = [self.arithmetic_job.uuid, 'aiida.in']
        result = self.cli_runner.invoke(command.calculation_inputcat, options)
        self.assertIsNone(result.exception)
        self.assertEquals(len(get_result_lines(result)), 1)
        self.assertEquals(get_result_lines(result)[0], '2 3')

    def test_calculation_outputcat(self):
        """Test verdi calculation outputcat"""
        options = []
        result = self.cli_runner.invoke(command.calculation_outputcat, options)
        self.assertIsNotNone(result.exception)

        options = [self.arithmetic_job.uuid]
        result = self.cli_runner.invoke(command.calculation_outputcat, options)
        self.assertIsNone(result.exception)
        self.assertEquals(len(get_result_lines(result)), 1)
        self.assertEquals(get_result_lines(result)[0], '5')

        options = [self.arithmetic_job.uuid, 'aiida.out']
        result = self.cli_runner.invoke(command.calculation_outputcat, options)
        self.assertIsNone(result.exception)
        self.assertEquals(len(get_result_lines(result)), 1)
        self.assertEquals(get_result_lines(result)[0], '5')

    def test_calculation_cleanworkdir(self):
        """Test verdi calculation cleanworkdir"""

        # Specifying no filtering options and no explicit calculations should exit with non-zero status
        options = []
        result = self.cli_runner.invoke(command.calculation_cleanworkdir, options)
        self.assertIsNotNone(result.exception)

        # Cannot specify both -p and -o options
        for flag_p in ['-p', '--past-days']:
            for flag_o in ['-o', '--older-than']:
                options = [flag_p, '5', flag_o, '1']
                result = self.cli_runner.invoke(command.calculation_cleanworkdir, options)
                self.assertIsNotNone(result.exception)

        # Without the force flag it should fail
        options = [str(self.result_job.uuid)]
        result = self.cli_runner.invoke(command.calculation_cleanworkdir, options)
        self.assertIsNotNone(result.exception)

        # With force flag we should find one calculation
        options = ['-f', str(self.result_job.uuid)]
        result = self.cli_runner.invoke(command.calculation_cleanworkdir, options)
        self.assertIsNone(result.exception)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi database` commands."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.utils import echo


@verdi.group('database')
def verdi_database():
    """Inspect and manage the database."""
    pass


@verdi_database.group('integrity')
def verdi_database_integrity():
    """Various commands that will check the integrity of the database and fix potential issues when asked."""
    pass


@verdi_database_integrity.command('duplicate-node-uuid')
@click.option(
    '-a',
    '--apply-patch',
    is_flag=True,
    help='Apply the proposed changes. If this flag is not passed, a dry run is performed instead.')
def database_integrity(apply_patch):
    """Detect and solve nodes with duplicate UUIDs.

    Before aiida-core v1.0.0, there was no uniqueness constraint on the UUID column of the Node table in the database.
    This made it possible to store multiple nodes with identical UUIDs without the database complaining. This was bug
    was fixed in aiida-core=1.0.0 by putting an explicit uniqueness constraint on node UUIDs on the database level.
    However, this would leave databases created before this patch with duplicate UUIDs in an inconsistent state. This
    command will run an analysis to detect duplicate UUIDs in the node table and solve it by generating new UUIDs. Note
    that it will not delete or merge any nodes.
    """
    from aiida.backends import settings
    from aiida.backends.utils import _load_dbenv_noschemacheck
    from aiida.common.setup import get_default_profile_name
    from aiida.manage.database.integrity import deduplicate_node_uuids

    if settings.AIIDADB_PROFILE is not None:
        profile = settings.AIIDADB_PROFILE
    else:
        profile = get_default_profile_name()

    _load_dbenv_noschemacheck(profile)

    try:
        messages = deduplicate_node_uuids(dry_run=not apply_patch)
    except Exception as exception:  # pylint: disable=broad-except
        echo.echo_critical('integrity check failed: {}'.format(str(exception)))
    else:
        for message in messages:
            echo.echo_info(message)

        if apply_patch:
            echo.echo_success('integrity patch completed')
        else:
            echo.echo_success('dry-run of integrity patch completed')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=too-many-arguments
"""`verdi process` command."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import click
from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import arguments, options
from aiida.cmdline.utils import decorators, echo
from aiida.cmdline.utils.query.calculation import CalculationQueryBuilder


@verdi.group('process')
def verdi_process():
    """Inspect and manage processes."""
    pass


@verdi_process.command('list')
@options.PROJECT(
    type=click.Choice(CalculationQueryBuilder.valid_projections), default=CalculationQueryBuilder.default_projections)
@options.ALL(help='Show all entries, regardless of their process state.')
@options.PROCESS_STATE()
@options.EXIT_STATUS()
@options.FAILED()
@options.PAST_DAYS()
@options.LIMIT()
@options.RAW()
@decorators.with_dbenv()
def process_list(all_entries, process_state, exit_status, failed, past_days, limit, project, raw):
    """Show a list of processes that are still running."""
    from tabulate import tabulate
    from aiida.cmdline.utils.common import print_last_process_state_change

    builder = CalculationQueryBuilder()
    filters = builder.get_filters(all_entries, process_state, exit_status, failed)
    query_set = builder.get_query_set(filters=filters, past_days=past_days, limit=limit)
    projected = builder.get_projected(query_set, projections=project)

    headers = projected.pop(0)

    if raw:
        tabulated = tabulate(projected, tablefmt='plain')
        echo.echo(tabulated)
    else:
        tabulated = tabulate(projected, headers=headers)
        echo.echo(tabulated)
        echo.echo('\nTotal results: {}\n'.format(len(projected)))
        print_last_process_state_change()


@verdi_process.command('kill')
@arguments.PROCESSES()
@options.TIMEOUT()
@decorators.with_dbenv()
@decorators.only_if_daemon_running(echo.echo_warning, 'daemon is not running, so process may not be reachable')
def process_kill(processes, timeout):
    """Kill running processes."""
    from aiida.work import RemoteException, DeliveryFailed, CommunicationTimeout, new_blocking_control_panel

    with new_blocking_control_panel(timeout=timeout) as control_panel:
        for process in processes:

            if process.is_terminated:
                echo.echo_error('Process<{}> is already terminated'.format(process.pk))
                continue

            try:
                if control_panel.kill_process(process.pk, msg='Killed through `verdi process kill`'):
                    echo.echo_success('killed Process<{}>'.format(process.pk))
                else:
                    echo.echo_error('problem killing Process<{}>'.format(process.pk))
            except CommunicationTimeout:
                echo.echo_error('call to kill Process<{}> timed out'.format(process.pk))
            except (RemoteException, DeliveryFailed) as exception:
                echo.echo_error('failed to kill Process<{}>: {}'.format(process.pk, exception))


@verdi_process.command('pause')
@arguments.PROCESSES()
@options.TIMEOUT()
@decorators.with_dbenv()
@decorators.only_if_daemon_running(echo.echo_warning, 'daemon is not running, so process may not be reachable')
def process_pause(processes, timeout):
    """Pause running processes."""
    from aiida.work import RemoteException, DeliveryFailed, CommunicationTimeout, new_blocking_control_panel

    with new_blocking_control_panel(timeout=timeout) as control_panel:
        for process in processes:

            if process.is_terminated:
                echo.echo_error('Process<{}> is already terminated'.format(process.pk))
                continue

            try:
                if control_panel.pause_process(process.pk, msg='Paused through `verdi process pause`'):
                    echo.echo_success('paused Process<{}>'.format(process.pk))
                else:
                    echo.echo_error('problem pausing Process<{}>'.format(process.pk))
            except CommunicationTimeout:
                echo.echo_error('call to pause Process<{}> timed out'.format(process.pk))
            except (RemoteException, DeliveryFailed) as exception:
                echo.echo_error('failed to pause Process<{}>: {}'.format(process.pk, exception))


@verdi_process.command('play')
@arguments.PROCESSES()
@options.TIMEOUT()
@decorators.with_dbenv()
@decorators.only_if_daemon_running(echo.echo_warning, 'daemon is not running, so process may not be reachable')
def process_play(processes, timeout):
    """Play paused processes."""
    from aiida.work import RemoteException, DeliveryFailed, CommunicationTimeout, new_blocking_control_panel

    with new_blocking_control_panel(timeout=timeout) as control_panel:
        for process in processes:

            if process.is_terminated:
                echo.echo_error('Process<{}> is already terminated'.format(process.pk))
                continue

            try:
                if control_panel.play_process(process.pk):
                    echo.echo_success('played Process<{}>'.format(process.pk))
                else:
                    echo.echo_critical('problem playing Process<{}>'.format(process.pk))
            except CommunicationTimeout:
                echo.echo_error('call to play Process<{}> timed out'.format(process.pk))
            except (RemoteException, DeliveryFailed) as exception:
                echo.echo_critical('failed to play Process<{}>: {}'.format(process.pk, exception))


@verdi_process.command('watch')
@arguments.PROCESSES()
@decorators.with_dbenv()
@decorators.only_if_daemon_running(echo.echo_warning, 'daemon is not running, so process may not be reachable')
def process_watch(processes):
    """Watch the state transitions for a process."""
    from kiwipy import BroadcastFilter
    from aiida.work.rmq import create_communicator

    def _print(body, sender, subject, correlation_id):
        echo.echo('pk={}, subject={}, body={}, correlation_id={}'.format(sender, subject, body, correlation_id))

    communicator = create_communicator()

    for process in processes:

        if process.is_terminated:
            echo.echo_error('Process<{}> is already terminated'.format(process.pk))
            continue

        communicator.add_broadcast_subscriber(BroadcastFilter(_print, sender=process.pk))

    try:
        communicator.await()
    except (SystemExit, KeyboardInterrupt):

        try:
            communicator.disconnect()
        except RuntimeError:
            pass

        # Reraise to trigger clicks builtin abort sequence
        raise



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from aiida.tools.dbimporters import DbImporter, DbImporterFactory
from aiida.tools.data.array.kpoints import get_kpoints_path, get_explicit_kpoints_path
from aiida.tools.data.structure import structure_to_spglib_tuple, spglib_tuple_to_structure



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=protected-access,invalid-name,too-many-arguments,too-many-locals
"""`verdi calculation` commands."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import os

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import arguments, options, types
from aiida.cmdline.utils import decorators, echo
from aiida.common.setup import get_property

LIST_CMDLINE_PROJECT_DEFAULT = get_property('verdishell.calculation_list')
LIST_CMDLINE_PROJECT_CHOICES = ('pk', 'state', 'ctime', 'job_state', 'calculation_state', 'scheduler_state', 'computer',
                                'type', 'description', 'label', 'uuid', 'mtime', 'user', 'sealed')


@verdi.group('calculation')
def verdi_calculation():
    """Inspect and manage calculations."""
    pass


@verdi_calculation.command('gotocomputer')
@arguments.CALCULATION(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:job', 'aiida.calculations:inline')))
def calculation_gotocomputer(calculation):
    """
    Open a shell and go to the calculation folder on the computer

    This command opens a ssh connection to the folder on the remote
    computer on which the calculation is being/has been executed.
    """
    from aiida.common.exceptions import NotExistent

    try:
        transport = calculation._get_transport()
    except NotExistent as exception:
        echo.echo_critical(exception)

    remote_workdir = calculation._get_remote_workdir()

    if not remote_workdir:
        echo.echo_critical('no remote work directory for this calculation, maybe the daemon did not submit it yet')

    command = transport.gotocomputer_command(remote_workdir)
    echo.echo_info('going to the remote work directory...')
    os.system(command)


@verdi_calculation.command('list')
@arguments.CALCULATIONS(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:job', 'aiida.calculations:inline')))
@options.CALCULATION_STATE()
@options.PROCESS_STATE()
@options.EXIT_STATUS()
@options.FAILED()
@options.PAST_DAYS()
@options.LIMIT()
@options.ORDER_BY()
@options.PROJECT(type=click.Choice(LIST_CMDLINE_PROJECT_CHOICES), default=LIST_CMDLINE_PROJECT_DEFAULT)
@options.GROUPS(help='Show only calculations that are contained within one or more of these groups.')
@options.ALL(help='Show all entries, regardless of their process state.')
@options.ALL_USERS()
@options.RAW()
@click.option(
    '-t',
    '--absolute-time',
    'absolute_time',
    is_flag=True,
    default=False,
    help='print the absolute creation time, rather than the relative creation time')
@decorators.with_dbenv()
def calculation_list(calculations, past_days, groups, all_entries, calculation_state, process_state, exit_status,
                     failed, limit, order_by, project, all_users, raw, absolute_time):
    """Return a list of job calculations that are still running."""
    from aiida.cmdline.utils.common import print_last_process_state_change
    from aiida.orm.calculation.job import JobCalculation
    from aiida.work.processes import ProcessState

    if all_entries:
        process_state = None
        calculation_state = None

    PROCESS_STATE_KEY = 'attributes.{}'.format(JobCalculation.PROCESS_STATE_KEY)
    EXIT_STATUS_KEY = 'attributes.{}'.format(JobCalculation.EXIT_STATUS_KEY)

    filters = {}

    if calculation_state:
        process_state = None

    if process_state:
        calculation_state = None
        filters[PROCESS_STATE_KEY] = {'in': process_state}

    if failed:
        calculation_state = None
        filters[PROCESS_STATE_KEY] = {'==': ProcessState.FINISHED.value}
        filters[EXIT_STATUS_KEY] = {'>': 0}

    if exit_status is not None:
        calculation_state = None
        filters[PROCESS_STATE_KEY] = {'==': ProcessState.FINISHED.value}
        filters[EXIT_STATUS_KEY] = {'==': exit_status}

    JobCalculation._list_calculations(
        states=calculation_state,
        past_days=past_days,
        pks=[calculation.pk for calculation in calculations],
        all_users=all_users,
        groups=groups,
        relative_ctime=not absolute_time,
        order_by=order_by,
        limit=limit,
        filters=filters,
        projections=project,
        raw=raw,
    )

    if not raw:
        print_last_process_state_change(process_type='calculation')


@verdi_calculation.command('res')
@arguments.CALCULATION(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:job', 'aiida.calculations:inline')))
@click.option('-f', '--format', 'fmt', type=click.STRING, default='json+date', help='format for the output')
@click.option('-k', '--keys', 'keys', type=click.STRING, cls=options.MultipleValueOption, help='show only these keys')
@decorators.with_dbenv()
def calculation_res(calculation, fmt, keys):
    """Print data from the result output node of a calculation."""
    from aiida.cmdline.utils.echo import echo_dictionary

    results = calculation.res._get_dict()

    if keys is not None:
        try:
            result = {k: results[k] for k in keys}
        except KeyError as exc:
            echo.echo_critical("key '{}' was not found in the .res dictionary".format(exc.args[0]))
    else:
        result = results

    echo_dictionary(result, fmt=fmt)


@verdi_calculation.command('show')
@arguments.CALCULATIONS(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:job', 'aiida.calculations:inline')))
@decorators.with_dbenv()
def calculation_show(calculations):
    """Show a summary for one or multiple calculations."""
    from aiida.cmdline.utils.common import get_node_info

    for calculation in calculations:
        echo.echo(get_node_info(calculation))


@verdi_calculation.command('logshow')
@arguments.CALCULATIONS(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:job', 'aiida.calculations:inline')))
@decorators.with_dbenv()
def calculation_logshow(calculations):
    """Show the log for one or multiple calculations."""
    from aiida.cmdline.utils.common import get_calculation_log_report

    for calculation in calculations:
        echo.echo(get_calculation_log_report(calculation))


@verdi_calculation.command('plugins')
@click.argument('entry_point', type=click.STRING, required=False)
@decorators.with_dbenv()
def calculation_plugins(entry_point):
    """Print a list of registered calculation plugins or details of a specific calculation plugin."""
    from aiida.common.exceptions import LoadingPluginFailed, MissingPluginError
    from aiida.plugins.entry_point import get_entry_point_names, load_entry_point
    import aiida.utils.json as json

    if entry_point:
        try:
            plugin = load_entry_point('aiida.calculations', entry_point)
        except (LoadingPluginFailed, MissingPluginError) as exception:
            echo.echo_critical(str(exception))
        else:
            echo.echo_info(entry_point)
            echo.echo_info(plugin.__doc__ if plugin.__doc__ else 'no docstring available')
            echo.echo(json.dumps(plugin.process().get_description(), indent=4))

    else:
        entry_points = get_entry_point_names('aiida.calculations')
        if entry_points:
            echo.echo('Registered calculation entry points:')
            for ep in entry_points:
                echo.echo("* {}".format(ep))

            echo.echo('')
            echo.echo_info('Pass the entry point as an argument to display detailed information')
        else:
            echo.echo_error('No calculation plugins found')


@verdi_calculation.command('inputcat')
@arguments.CALCULATION(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:job', 'aiida.calculations:inline')))
@click.argument('path', type=click.STRING, required=False)
@decorators.with_dbenv()
def calculation_inputcat(calculation, path):
    """
    Show the contents of a file with relative PATH in the raw input folder of the CALCULATION.

    If PATH is not specified, the default input file path will be used, if defined by the calculation plugin class.
    """
    from aiida.cmdline.utils.repository import cat_repo_files
    from aiida.plugins.entry_point import get_entry_point_from_class

    if path is None:

        path = calculation._DEFAULT_INPUT_FILE

        if path is None:
            cls = calculation.__class__
            _, entry_point = get_entry_point_from_class(cls.__module__, cls.__name__)
            echo.echo_critical('{} does not define a default input file. Please specify a path explicitly'.format(
                entry_point.name))

    try:
        cat_repo_files(calculation, os.path.join('raw_input', path))
    except ValueError as exc:
        echo.echo_critical(str(exc))
    except IOError as e:
        import errno
        # Ignore Broken pipe errors, re-raise everything else
        if e.errno == errno.EPIPE:
            pass
        else:
            raise


@verdi_calculation.command('outputcat')
@arguments.CALCULATION(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:job', 'aiida.calculations:inline')))
@click.argument('path', type=click.STRING, required=False)
@decorators.with_dbenv()
def calculation_outputcat(calculation, path):
    """
    Show the contents of a file with relative PATH in the retrieved folder of the CALCULATION.

    If PATH is not specified, the default output file path will be used, if defined by the calculation plugin class.
    Content can only be shown after the daemon has retrieved the remote files.
    """
    from aiida.cmdline.utils.repository import cat_repo_files
    from aiida.plugins.entry_point import get_entry_point_from_class

    if path is None:

        path = calculation._DEFAULT_OUTPUT_FILE

        if path is None:
            cls = calculation.__class__
            _, entry_point = get_entry_point_from_class(cls.__module__, cls.__name__)
            echo.echo_critical('{} does not define a default output file. Please specify a path explicitly'.format(
                entry_point.name))

    try:
        retrieved = calculation.out.retrieved
    except AttributeError:
        echo.echo_critical("No 'retrieved' node found. Have the calculation files already been retrieved?")

    try:
        cat_repo_files(retrieved, os.path.join('path', path))
    except ValueError as exc:
        echo.echo_critical(str(exc))
    except IOError as e:
        import errno
        # Ignore Broken pipe errors, re-raise everything else
        if e.errno == errno.EPIPE:
            pass
        else:
            raise


@verdi_calculation.command('inputls')
@decorators.with_dbenv()
@arguments.CALCULATION(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:job', 'aiida.calculations:inline')))
@click.argument('path', type=click.STRING, required=False)
@click.option('-c', '--color', 'color', is_flag=True, default=False, help='color folders with a different color')
def calculation_inputls(calculation, path, color):
    """
    Show the list of files in the directory with relative PATH in the raw input folder of the CALCULATION.

    If PATH is not specified, the base path of the input folder will be used.
    """
    from aiida.cmdline.utils.repository import list_repo_files
    from aiida.orm.implementation.general.calculation.job import _input_subfolder

    if path is not None:
        fullpath = os.path.join(_input_subfolder, path)
    else:
        fullpath = _input_subfolder

    try:
        list_repo_files(calculation, fullpath, color)
    except ValueError as exception:
        echo.echo_critical(exception)


@verdi_calculation.command('outputls')
@decorators.with_dbenv()
@arguments.CALCULATION(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:job', 'aiida.calculations:inline')))
@click.argument('path', type=click.STRING, required=False)
@click.option('-c', '--color', 'color', is_flag=True, default=False, help='color folders with a different color')
def calculation_outputls(calculation, path, color):
    """
    Show the list of files in the directory with relative PATH in the raw input folder of the CALCULATION.

    If PATH is not specified, the base path of the retrieved folder will be used.
    Content can only be showm after the daemon has retrieved the remote files.
    """
    from aiida.cmdline.utils.repository import list_repo_files

    if path is not None:
        fullpath = os.path.join(calculation._path_subfolder_name, path)
    else:
        fullpath = calculation._path_subfolder_name

    try:
        retrieved = calculation.out.retrieved
    except AttributeError:
        echo.echo_critical("No 'retrieved' node found. Have the calculation files already been retrieved?")

    try:
        list_repo_files(retrieved, fullpath, color)
    except ValueError as exception:
        echo.echo_critical(exception)


@verdi_calculation.command('kill')
@decorators.with_dbenv()
@arguments.CALCULATIONS(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:job', 'aiida.calculations:inline')))
@options.FORCE()
@decorators.deprecated_command("This command will be removed in a future release. Use 'verdi process kill' instead.")
def calculation_kill(calculations, force):
    """Kill one or multiple running calculations."""
    from aiida import work

    if not force:
        warning = 'Are you sure you want to kill {} calculations?'.format(len(calculations))
        click.confirm(warning, abort=True)

    with work.new_control_panel() as control_panel:
        futures = []
        for calculation in calculations:
            try:
                future = control_panel.kill_process(calculation)
                futures.append((calculation, future))
            except (work.RemoteException, work.DeliveryFailed) as exc:
                echo.echo_error('Calculation<{}> killing failed {}'.format(calculation, exc))

        for future in futures:
            result = control_panel._communicator.await(future[1])
            if result:
                echo.echo_success('Calculation<{}> successfully killed'.format(future[0]))
            else:
                echo.echo_error('Calculation<{}> killing failed {}'.format(future[0], result))


@verdi_calculation.command('cleanworkdir')
@decorators.with_dbenv()
@arguments.CALCULATIONS(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:job', 'aiida.calculations:inline')))
@options.PAST_DAYS(default=None)
@options.OLDER_THAN(default=None)
@options.COMPUTERS(help='include only calculations that were ran on these computers')
@options.FORCE()
def calculation_cleanworkdir(calculations, past_days, older_than, computers, force):
    """
    Clean all content of all output remote folders of calculations.

    If no explicit calculations are specified as arguments, one or both of the -p and -o options has to be specified.
    If both are specified, a logical AND is done between the two, i.e. the calculations that will be cleaned have been
    modified AFTER [-p option] days from now, but BEFORE [-o option] days from now.
    """
    from aiida.orm.backend import construct_backend
    from aiida.orm.utils.loaders import ComputerEntityLoader, IdentifierType
    from aiida.orm.utils.remote import clean_remote, get_calculation_remote_paths

    if calculations:
        if (past_days is not None and older_than is not None):
            echo.echo_critical('specify either explicit calculations or use the filtering options')
    else:
        if (past_days is None and older_than is None):
            echo.echo_critical('if no explicit calculations are specified, at least one filtering option is required')

    calculations_pks = [calculation.pk for calculation in calculations]
    path_mapping = get_calculation_remote_paths(calculations_pks, past_days, older_than, computers)

    if path_mapping is None:
        echo.echo_critical('no calculations found with the given criteria')

    if not force:
        path_count = sum([len(paths) for computer, paths in path_mapping.items()])
        warning = 'Are you sure you want to clean the work directory of {} calculations?'.format(path_count)
        click.confirm(warning, abort=True)

    backend = construct_backend()
    user = backend.users.get_automatic_user()

    for computer_uuid, paths in path_mapping.items():

        counter = 0
        computer = ComputerEntityLoader.load_entity(computer_uuid, identifier_type=IdentifierType.UUID)
        transport = backend.authinfos.get(computer, user).get_transport()

        with transport:
            for path in paths:
                clean_remote(transport, path)
                counter += 1

        echo.echo_success('{} remote folders cleaned on {}'.format(counter, computer.name))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=superfluous-parens
"""
This allows to manage comments from command line.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import arguments, options
from aiida.cmdline.utils import decorators, echo, multi_line_input


@verdi.group('comment')
def verdi_comment():
    """Inspect, create and manage comments."""
    pass


@verdi_comment.command()
@click.option('--comment', '-c', type=str, required=False)
@arguments.NODES(required=True)
@decorators.with_dbenv()
def add(comment, nodes):
    """
    Add comment to one or more nodes in the database
    """
    from aiida.orm.backend import construct_backend

    backend = construct_backend()
    user = backend.users.get_automatic_user()

    if not comment:
        comment = multi_line_input.edit_comment()

    for node in nodes:
        node.add_comment(comment, user)

    echo.echo_info("Comment added to node(s) '{}'".format(", ".join([str(node.pk) for node in nodes])))


@verdi_comment.command()
@options.USER()
@arguments.NODES()
@decorators.with_dbenv()
def show(user, nodes):
    """
    Show the comments of (a) node(s) in the database
    """
    for node in nodes:
        all_comments = node.get_comments()

        if user is not None:
            to_print = [i for i in all_comments if i['user__email'] == user.email]
            if not to_print:
                valid_users = ", ".join(set(["'" + i['user__email'] + "'" for i in all_comments]))
                echo.echo_info("Nothing found for user '{}'.\n"
                               "Valid users found for Node {} are: {}.".format(user, node.pk, valid_users))

        else:
            to_print = all_comments

        for i in to_print:
            comment_msg = [
                "***********************************************************", "Comment of '{}' on {}".format(
                    i['user__email'], i['ctime'].strftime("%Y-%m-%d %H:%M")), "PK {} ID {}. Last modified on {}".format(
                        node.pk, i['pk'], i['mtime'].strftime("%Y-%m-%d %H:%M")), "", "{}".format(i['content']), ""
            ]
            echo.echo_info("\n".join(comment_msg))

        # If there is nothing to print, print a message
        if not to_print:
            echo.echo_info("No comments found.")


@verdi_comment.command()
@click.option(
    '--all',
    '-a',
    'remove_all',
    default=False,
    is_flag=True,
    help='If used, deletes all the comments of the active user attached to the node')
@options.FORCE()
@arguments.NODE()
@click.argument('comment_id', type=int, required=False, metavar='COMMENT_ID')
@decorators.with_dbenv()
def remove(remove_all, force, node, comment_id):
    """
    Remove comment(s) of a node. The user can only remove their own comments.

    pk = The pk (an integer) of the node

    id = #ID of the comment to be removed from node #PK
    """
    # Note: in fact, the user can still manually delete any comment
    from aiida.orm.backend import construct_backend

    backend = construct_backend()
    user = backend.users.get_automatic_user()

    if comment_id is None and not remove_all:
        echo.echo_error("One argument between -a and ID must be provided")
        return 101

    if comment_id is not None and remove_all:
        echo.echo_error("Cannot use -a together with a comment id")
        return 102

    if remove_all:
        comment_id = None

    if not force:
        if remove_all:
            click.confirm("Delete all comments of user {} on node <{}>? ".format(user, node.pk), abort=True)
        else:
            click.confirm("Delete comment? ", abort=True)

    comments = node.get_comment_obj(comment_id=comment_id, user=user)
    for comment in comments:
        comment.delete()
    echo.echo_info("Deleted {} comments.".format(len(comments)))

    return 0


@verdi_comment.command()
@click.option('--comment', '-c', type=str, required=False)
@arguments.NODE()
@click.argument('comment_id', type=int, metavar='COMMENT_ID')
@decorators.with_dbenv()
def update(comment, node, comment_id):
    """
    Update a comment.

    id      = The id of the comment
    comment = The comment (a string) to be added to the node(s)
    """
    from aiida.orm.backend import construct_backend
    backend = construct_backend()
    user = backend.users.get_automatic_user()

    # read the comment from terminal if it is not on command line
    if comment is None:
        try:
            current_comment = node.get_comments(comment_id)[0]
        except IndexError:
            echo.echo_error("Comment with id '{}' not found".format(comment_id))
            return 1

        comment = multi_line_input.edit_comment(current_comment['content'])

    # pylint: disable=protected-access
    node._update_comment(comment, comment_id, user)

    return 0



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""A utility module with a factory of standard QueryBuilder instances for Calculation nodes."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from aiida.common.utils import classproperty
from aiida.cmdline.utils.query.mapping import CalculationProjectionMapper


class CalculationQueryBuilder(object):
    """Utility class to construct a QueryBuilder instance for Calculation nodes and project the query set."""

    _default_projections = ('pk', 'ctime', 'state', 'process_label', 'process_status')
    _valid_projections = ('pk', 'uuid', 'ctime', 'mtime', 'state', 'process_state', 'process_status', 'exit_status',
                          'sealed', 'process_label', 'label', 'description', 'type', 'paused', 'process_type')

    def __init__(self, mapper=None):
        if mapper is None:
            self._mapper = CalculationProjectionMapper(self._valid_projections)
        else:
            self._mapper = mapper

    @property
    def mapper(self):
        return self._mapper

    @classproperty
    def default_projections(self):
        return self._default_projections

    @classproperty
    def valid_projections(self):
        return self._valid_projections

    def get_filters(self, all_entries=False, process_state=None, exit_status=None, failed=False, node_types=None):
        """
        Return a set of QueryBuilder filters based on typical command line options.

        :param node_types: a tuple of node classes to filter for (must be sub classes of Calculation)
        :param all_entries: boolean to negate filtering for process state
        :param process_state: filter for this process state
        :param exit_status: filter for this exit status
        :param failed: boolean to filter only failed processes
        :return: dictionary of filters suitable for a QueryBuilder.append() call
        """
        from aiida.work import ProcessState

        exit_status_attribute = self.mapper.get_attribute('exit_status')
        process_state_attribute = self.mapper.get_attribute('process_state')

        filters = {}

        if node_types is not None:
            filters['or'] = []
            for node_type in node_types:
                filters['or'].append({'type': node_type.plugin_type_string})

        if process_state and not all_entries:
            filters[process_state_attribute] = {'in': process_state}

        if failed:
            filters[process_state_attribute] = {'==': ProcessState.FINISHED.value}
            filters[exit_status_attribute] = {'>': 0}

        if exit_status is not None:
            filters[process_state_attribute] = {'==': ProcessState.FINISHED.value}
            filters[exit_status_attribute] = {'==': exit_status}

        return filters

    def get_query_set(self, filters=None, order_by=None, past_days=None, limit=None):
        """
        Return the query set of calculations for the given filters and query parameters

        :param filters: rules to filter query results with
        :param order_by: order the query set by this criterion
        :param past_days: only include entries from the last past days
        :param limit: limit the query set to this number of entries
        :return: the query set, a list of dictionaries
        """
        import datetime

        from aiida.orm.calculation import Calculation
        from aiida.orm.querybuilder import QueryBuilder
        from aiida.utils import timezone

        projected_attributes = [self.mapper.get_attribute(projection) for projection in self._valid_projections]

        if filters is None:
            filters = {}

        if past_days is not None:
            filters['ctime'] = {'>': timezone.now() - datetime.timedelta(days=past_days)}

        builder = QueryBuilder()
        builder.append(cls=Calculation, filters=filters, project=projected_attributes, tag='calculation')

        if order_by is not None:
            builder.order_by({'calculation': order_by})
        else:
            builder.order_by({'calculation': {'ctime': 'asc'}})

        if limit is not None:
            builder.limit(limit)

        return builder.iterdict()

    def get_projected(self, query_set, projections):
        """
        Project the query set for the given set of projections
        """
        header = [self.mapper.get_label(projection) for projection in projections]
        result = [header]

        for query_result in query_set:
            result_row = [self.mapper.format(projection, query_result['calculation']) for projection in projections]
            result.append(result_row)

        return result



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from aiida.orm.backend import Backend

from aiida.backends.djsite.queries import DjangoQueryManager
from . import authinfo
from . import computer
from . import log
from . import user


class DjangoBackend(Backend):

    def __init__(self):
        self._logs = log.DjangoLogCollection(self)
        self._users = user.DjangoUserCollection(self)
        self._authinfos = authinfo.DjangoAuthInfoCollection(self)
        self._computers = computer.DjangoComputerCollection(self)
        self._query_manager = DjangoQueryManager(self)

    @property
    def logs(self):
        return self._logs

    @property
    def users(self):
        return self._users

    @property
    def authinfos(self):
        return self._authinfos

    @property
    def computers(self):
        return self._computers

    @property
    def query_manager(self):
        return self._query_manager

    def query_builder(self):
        # For now this doesn't select the Django one because that's
        # done in the constructor of QueryBuilder but ideally this
        # should be reworked so the Django version subclasses QueryBuilder
        from aiida.orm.querybuilder import QueryBuilder
        return QueryBuilder()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from aiida.backends.testbase import AiidaTestCase
from aiida.cmdline.params.types import LegacyWorkflowParamType


class TestLegacyWorkflowParamType(AiidaTestCase):

    @classmethod
    def setUpClass(cls):
        from aiida.workflows.test import WFTestEmpty, WFTestSimpleWithSubWF

        super(TestLegacyWorkflowParamType, cls).setUpClass()

        cls.workflow = WFTestEmpty()
        cls.workflow.label = 'Unique Label'
        cls.workflow.store()
        cls.wf_type = LegacyWorkflowParamType()

    def test_get_by_id(self):
        identifier = str(self.workflow.pk)
        self.assertEqual(self.wf_type.convert(identifier, None, None).uuid, self.workflow.uuid)

    def test_get_by_uuid(self):
        identifier = str(self.workflow.uuid)
        self.assertEqual(self.wf_type.convert(identifier, None, None).uuid, self.workflow.uuid)

    def test_get_by_label(self):
        identifier = str(self.workflow.label)
        self.assertEqual(self.wf_type.convert(identifier, None, None).uuid, self.workflow.uuid)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import functools
import logging
import shutil
import sys
import tempfile

import six
from tornado.gen import coroutine, Return

import plumpy
from plumpy.ports import PortNamespace
from aiida.common.datastructures import calc_states, is_progressive_state_change
from aiida.common.exceptions import TransportTaskException
from aiida.common import exceptions
from aiida.common.lang import override
from aiida.daemon import execmanager
from aiida.orm.calculation.job import JobCalculation
from aiida.scheduler.datastructures import JOB_STATES
from aiida.work.process_builder import JobProcessBuilder
from aiida.work.utils import exponential_backoff_retry, interruptable_task

from . import persistence
from . import processes

__all__ = ['JobProcess']

UPLOAD_COMMAND = 'upload'
SUBMIT_COMMAND = 'submit'
UPDATE_COMMAND = 'update'
RETRIEVE_COMMAND = 'retrieve'
KILL_COMMAND = 'kill'

TRANSPORT_TASK_RETRY_INITIAL_INTERVAL = 20
TRANSPORT_TASK_MAXIMUM_ATTEMTPS = 5

logger = logging.getLogger(__name__)


@coroutine
def task_upload_job(node, transport_queue, calc_info, script_filename, cancellable):
    """
    Transport task that will attempt to upload the files of a job calculation to the remote

    The task will first request a transport from the queue. Once the transport is yielded, the relevant execmanager
    function is called, wrapped in the exponential_backoff_retry coroutine, which, in case of a caught exception, will
    retry after an interval that increases exponentially with the number of retries, for a maximum number of retries.
    If all retries fail, the task will raise a TransportTaskException

    :param node: the node that represents the job calculation
    :param transport_queue: the TransportQueue from which to request a Transport
    :param calc_info: the calculation info datastructure returned by `JobCalculation._presubmit`
    :param script_filename: the job launch script returned by `JobCalculation._presubmit`
    :param cancellable: the cancelled flag that will be queried to determine whether the task was cancelled
    :type cancellable: :class:`aiida.work.utils.InterruptableFuture`
    :raises: Return if the tasks was successfully completed
    :raises: TransportTaskException if after the maximum number of retries the transport task still excepted
    """
    initial_interval = TRANSPORT_TASK_RETRY_INITIAL_INTERVAL
    max_attempts = TRANSPORT_TASK_MAXIMUM_ATTEMTPS

    authinfo = node.get_computer().get_authinfo(node.get_user())

    state_pending = calc_states.SUBMITTING

    if is_progressive_state_change(node.get_state(), state_pending):
        node._set_state(state_pending)
    else:
        logger.warning('ignored invalid proposed state change: {} to {}'.format(node.get_state(), state_pending))

    @coroutine
    def do_upload():
        with transport_queue.request_transport(authinfo) as request:
            transport = yield cancellable.with_interrupt(request)

            logger.info('uploading calculation<{}>'.format(node.pk))
            raise Return(execmanager.upload_calculation(node, transport, calc_info, script_filename))

    try:
        result = yield exponential_backoff_retry(
            do_upload, initial_interval, max_attempts, logger=node.logger, ignore_exceptions=plumpy.CancelledError)
    except plumpy.CancelledError:
        pass
    except Exception:
        logger.warning('uploading calculation<{}> failed'.format(node.pk))
        raise TransportTaskException('upload_calculation failed {} times consecutively'.format(max_attempts))
    else:
        logger.info('uploading calculation<{}> successful'.format(node.pk))
        raise Return(result)


@coroutine
def task_submit_job(node, transport_queue, calc_info, script_filename, cancellable):
    """
    Transport task that will attempt to submit a job calculation

    The task will first request a transport from the queue. Once the transport is yielded, the relevant execmanager
    function is called, wrapped in the exponential_backoff_retry coroutine, which, in case of a caught exception, will
    retry after an interval that increases exponentially with the number of retries, for a maximum number of retries.
    If all retries fail, the task will raise a TransportTaskException

    :param node: the node that represents the job calculation
    :param transport_queue: the TransportQueue from which to request a Transport
    :param calc_info: the calculation info datastructure returned by `JobCalculation._presubmit`
    :param script_filename: the job launch script returned by `JobCalculation._presubmit`
    :param cancellable: the cancelled flag that will be queried to determine whether the task was cancelled
    :type cancellable: :class:`aiida.work.utils.InterruptableFuture`
    :raises: Return if the tasks was successfully completed
    :raises: TransportTaskException if after the maximum number of retries the transport task still excepted
    """
    if node.get_state() == calc_states.WITHSCHEDULER:
        assert node.get_job_id() is not None, 'job is WITHSCHEDULER, however, it does not have a job id'
        logger.warning('calculation<{}> already marked as WITHSCHEDULER, skipping task_submit_job'.format(node.pk))
        raise Return(node.get_job_id())

    initial_interval = TRANSPORT_TASK_RETRY_INITIAL_INTERVAL
    max_attempts = TRANSPORT_TASK_MAXIMUM_ATTEMTPS

    authinfo = node.get_computer().get_authinfo(node.get_user())

    @coroutine
    def do_submit():
        with transport_queue.request_transport(authinfo) as request:
            transport = yield cancellable.with_interrupt(request)

            logger.info('submitting calculation<{}>'.format(node.pk))
            raise Return(execmanager.submit_calculation(node, transport, calc_info, script_filename))

    try:
        result = yield exponential_backoff_retry(
            do_submit, initial_interval, max_attempts, logger=node.logger, ignore_exceptions=plumpy.Interruption)
    except plumpy.Interruption:
        pass
    except Exception:
        logger.warning('submitting calculation<{}> failed'.format(node.pk))
        raise TransportTaskException('submit_calculation failed {} times consecutively'.format(max_attempts))
    else:
        logger.info('submitting calculation<{}> successful'.format(node.pk))
        node._set_state(calc_states.WITHSCHEDULER)
        raise Return(result)


@coroutine
def task_update_job(node, job_manager, cancellable):
    """
    Transport task that will attempt to update the scheduler status of the job calculation

    The task will first request a transport from the queue. Once the transport is yielded, the relevant execmanager
    function is called, wrapped in the exponential_backoff_retry coroutine, which, in case of a caught exception, will
    retry after an interval that increases exponentially with the number of retries, for a maximum number of retries.
    If all retries fail, the task will raise a TransportTaskException

    :param node: the node that represents the job calculation
    :type node: :class:`aiida.orm.calculation.JobCalculation`
    :param job_manager: The job manager
    :type job_manager: :class:`aiida.work.job_calcs.JobManager`
    :param cancellable: A cancel flag
    :type cancellable: :class:`aiida.work.utils.InterruptableFuture`
    :raises: Return containing True if the tasks was successfully completed, False otherwise
    """
    if node.get_state() == calc_states.COMPUTED:
        logger.warning('calculation<{}> already marked as COMPUTED, skipping task_update_job'.format(node.pk))
        raise Return(True)

    initial_interval = TRANSPORT_TASK_RETRY_INITIAL_INTERVAL
    max_attempts = TRANSPORT_TASK_MAXIMUM_ATTEMTPS

    authinfo = node.get_computer().get_authinfo(node.get_user())
    job_id = node.get_job_id()

    @coroutine
    def do_update():
        # Get the update request
        with job_manager.request_job_info_update(authinfo, job_id) as update_request:
            job_info = yield cancellable.with_interrupt(update_request)

        if job_info is None:
            # If the job is computed or not found assume it's done
            node._set_scheduler_state(JOB_STATES.DONE)
            job_done = True
        else:
            node._set_last_jobinfo(job_info)
            node._set_scheduler_state(job_info.job_state)
            job_done = job_info.job_state == JOB_STATES.DONE

        raise Return(job_done)

    try:
        job_done = yield exponential_backoff_retry(
            do_update, initial_interval, max_attempts, logger=node.logger, ignore_exceptions=plumpy.Interruption)
    except plumpy.Interruption:
        raise
    except Exception:
        logger.warning('updating calculation<{}> failed'.format(node.pk))
        raise TransportTaskException('update_calculation failed {} times consecutively'.format(max_attempts))
    else:
        logger.info('updating calculation<{}> successful'.format(node.pk))
        if job_done:
            node._set_state(calc_states.COMPUTED)

        raise Return(job_done)


@coroutine
def task_retrieve_job(node, transport_queue, retrieved_temporary_folder, cancellable):
    """
    Transport task that will attempt to retrieve all files of a completed job calculation

    The task will first request a transport from the queue. Once the transport is yielded, the relevant execmanager
    function is called, wrapped in the exponential_backoff_retry coroutine, which, in case of a caught exception, will
    retry after an interval that increases exponentially with the number of retries, for a maximum number of retries.
    If all retries fail, the task will raise a TransportTaskException

    :param node: the node that represents the job calculation
    :param transport_queue: the TransportQueue from which to request a Transport
    :param cancellable: the cancelled flag that will be queried to determine whether the task was cancelled
    :type cancellable: :class:`aiida.work.utils.InterruptableFuture`
    :raises: Return if the tasks was successfully completed
    :raises: TransportTaskException if after the maximum number of retries the transport task still excepted
    """
    if node.get_state() == calc_states.PARSING:
        logger.warning('calculation<{}> already marked as PARSING, skipping task_retrieve_job'.format(node.pk))
        raise Return(True)

    initial_interval = TRANSPORT_TASK_RETRY_INITIAL_INTERVAL
    max_attempts = TRANSPORT_TASK_MAXIMUM_ATTEMTPS

    authinfo = node.get_computer().get_authinfo(node.get_user())

    @coroutine
    def do_retrieve():
        with transport_queue.request_transport(authinfo) as request:
            transport = yield cancellable.with_interrupt(request)

            logger.info('retrieving calculation<{}>'.format(node.pk))
            raise Return(execmanager.retrieve_calculation(node, transport, retrieved_temporary_folder))

    state_pending = calc_states.RETRIEVING

    if is_progressive_state_change(node.get_state(), state_pending):
        node._set_state(state_pending)
    else:
        logger.warning('ignored invalid proposed state change: {} to {}'.format(node.get_state(), state_pending))

    try:
        result = yield exponential_backoff_retry(
            do_retrieve, initial_interval, max_attempts, logger=node.logger, ignore_exceptions=plumpy.Interruption)
    except plumpy.Interruption:
        raise
    except Exception:
        logger.warning('retrieving calculation<{}> failed'.format(node.pk))
        raise TransportTaskException('retrieve_calculation failed {} times consecutively'.format(max_attempts))
    else:
        node._set_state(calc_states.PARSING)
        logger.info('retrieving calculation<{}> successful'.format(node.pk))
        raise Return(result)


@coroutine
def task_kill_job(node, transport_queue, cancellable):
    """
    Transport task that will attempt to kill a job calculation

    The task will first request a transport from the queue. Once the transport is yielded, the relevant execmanager
    function is called, wrapped in the exponential_backoff_retry coroutine, which, in case of a caught exception, will
    retry after an interval that increases exponentially with the number of retries, for a maximum number of retries.
    If all retries fail, the task will raise a TransportTaskException

    :param node: the node that represents the job calculation
    :param transport_queue: the TransportQueue from which to request a Transport
    :param cancellable: the cancelled flag that will be queried to determine whether the task was cancelled
    :type cancellable: :class:`aiida.work.utils.InterruptableFuture`
    :raises: Return if the tasks was successfully completed
    :raises: TransportTaskException if after the maximum number of retries the transport task still excepted
    """
    initial_interval = TRANSPORT_TASK_RETRY_INITIAL_INTERVAL
    max_attempts = TRANSPORT_TASK_MAXIMUM_ATTEMTPS

    if node.get_state() in [calc_states.NEW, calc_states.TOSUBMIT, calc_states.SUBMITTING]:
        logger.warning('calculation<{}> killed, it was in the {} state'.format(node.pk, node.get_state()))
        raise Return(True)

    authinfo = node.get_computer().get_authinfo(node.get_user())

    @coroutine
    def do_kill():
        with transport_queue.request_transport(authinfo) as request:
            transport = yield cancellable.with_interrupt(request)
            logger.info('killing calculation<{}>'.format(node.pk))
            raise Return(execmanager.kill_calculation(node, transport))

    try:
        result = yield exponential_backoff_retry(do_kill, initial_interval, max_attempts, logger=node.logger)
    except plumpy.Interruption:
        raise
    except Exception:
        logger.warning('killing calculation<{}> failed'.format(node.pk))
        raise TransportTaskException('kill_calculation failed {} times consecutively'.format(max_attempts))
    else:
        logger.info('killing calculation<{}> successful'.format(node.pk))
        node._set_scheduler_state(JOB_STATES.DONE)
        raise Return(result)


class Waiting(plumpy.Waiting):
    """
    The waiting state for the JobCalculation.
    """

    def __init__(self, process, done_callback, msg=None, data=None):
        super(Waiting, self).__init__(process, done_callback, msg, data)
        self._task = None
        self._killing = None

    def load_instance_state(self, saved_state, load_context):
        super(Waiting, self).load_instance_state(saved_state, load_context)
        self._task = None
        self._killing = None

    @coroutine
    def execute(self):

        calculation = self.process.calc
        transport_queue = self.process.runner.transport

        if isinstance(self.data, tuple):
            command = self.data[0]
            args = self.data[1:]
        else:
            command = self.data

        calculation._set_process_status('Waiting for transport task: {}'.format(command))

        try:

            if command == UPLOAD_COMMAND:
                calc_info, script_filename = yield self._launch_task(task_upload_job, calculation, transport_queue,
                                                                     *args)
                raise Return(self.submit(calc_info, script_filename))

            elif command == SUBMIT_COMMAND:
                yield self._launch_task(task_submit_job, calculation, transport_queue, *args)
                raise Return(self.scheduler_update())

            elif self.data == UPDATE_COMMAND:
                job_done = False

                while not job_done:
                    job_done = yield self._launch_task(task_update_job, calculation,
                                                       self.process.runner.job_manager)

                raise Return(self.retrieve())

            elif self.data == RETRIEVE_COMMAND:
                # Create a temporary folder that has to be deleted by JobProcess.retrieved after successful parsing
                temp_folder = tempfile.mkdtemp()
                yield self._launch_task(task_retrieve_job, calculation, transport_queue, temp_folder)
                raise Return(self.retrieved(temp_folder))

            else:
                raise RuntimeError('Unknown waiting command')

        except TransportTaskException as exception:
            raise plumpy.PauseInterruption('Pausing after failed transport task: {}'.format(exception))
        except plumpy.KillInterruption:
            exc_info = sys.exc_info()
            yield self._launch_task(task_kill_job, calculation, transport_queue)
            self._killing.set_result(True)
            six.reraise(*exc_info)
        except Return:
            calculation._set_process_status(None)
            raise
        except (plumpy.Interruption, plumpy.CancelledError):
            calculation._set_process_status('Transport task {} was interrupted'.format(command))
            raise
        finally:
            # If we were trying to kill but we didn't deal with it, make sure it's set here
            if self._killing and not self._killing.done():
                self._killing.set_result(False)

    @coroutine
    def _launch_task(self, coro, *args, **kwargs):
        task_fn = functools.partial(coro, *args, **kwargs)
        try:
            self._task = interruptable_task(task_fn)
            result = yield self._task
            raise Return(result)
        finally:
            self._task = None

    def upload(self, calc_info, script_filename):
        """
        Create the next state to go to

        :return: The appropriate WAITING state
        """
        return self.create_state(
            processes.ProcessState.WAITING,
            None,
            msg='Waiting for calculation folder upload',
            data=(UPLOAD_COMMAND, calc_info, script_filename))

    def submit(self, calc_info, script_filename):
        """
        Create the next state to go to

        :return: The appropriate WAITING state
        """
        return self.create_state(
            processes.ProcessState.WAITING,
            None,
            msg='Waiting for scheduler submission',
            data=(SUBMIT_COMMAND, calc_info, script_filename))

    def scheduler_update(self):
        """
        Create the next state to go to

        :return: The appropriate WAITING state
        """
        return self.create_state(
            processes.ProcessState.WAITING,
            None,
            msg='Waiting for scheduler update',
            data=UPDATE_COMMAND)

    def retrieve(self):
        """
        Create the next state to go to in order to retrieve

        :return: The appropriate WAITING state
        """
        return self.create_state(
            processes.ProcessState.WAITING,
            None,
            msg='Waiting to retrieve',
            data=RETRIEVE_COMMAND)

    def retrieved(self, retrieved_temporary_folder):
        """
        Create the next state to go to after retrieving
        :param retrieved_temporary_folder: The temporary folder used in retrieving, this will
            be used in parsing.
        :return: The appropriate RUNNING state
        """
        return self.create_state(
            processes.ProcessState.RUNNING,
            self.process.retrieved,
            retrieved_temporary_folder)

    def interrupt(self, reason):
        """Interrupt the Waiting state by calling interrupt on the transport task InterruptableFuture."""
        if self._task is not None:
            self._task.interrupt(reason)

        if isinstance(reason, plumpy.KillInterruption):
            if self._killing is None:
                self._killing = plumpy.Future()
            return self._killing


class JobProcess(processes.Process):
    TRANSPORT_OPERATION = 'TRANSPORT_OPERATION'
    CALC_NODE_LABEL = 'calc_node'
    OPTIONS_INPUT_LABEL = 'options'

    _calc_class = None

    @classmethod
    def get_builder(cls):
        return JobProcessBuilder(cls)

    @classmethod
    def build(cls, calc_class):
        from aiida.orm.data import Data

        def define(cls_, spec):
            super(JobProcess, cls_).define(spec)

            spec.input_namespace(cls.OPTIONS_INPUT_LABEL, help='various options')
            for key, option in calc_class.options.items():
                spec.input(
                    '{}.{}'.format(cls.OPTIONS_INPUT_LABEL, key),
                    required=option.get('required', True),
                    valid_type=option.get('valid_type', object),  # Should match everything, as in all types are valid
                    non_db=option.get('non_db', True),
                    help=option.get('help', '')
                )

            # Define the actual inputs based on the use methods of the calculation class
            for key, use_method in calc_class._use_methods.items():

                valid_type = use_method['valid_types']
                docstring = use_method.get('docstring', None)
                additional_parameter = use_method.get('additional_parameter')

                if additional_parameter:
                    spec.input_namespace(key, help=docstring, valid_type=valid_type, required=False, dynamic=True)
                else:
                    spec.input(key, help=docstring, valid_type=valid_type, required=False)

            # Outputs
            spec.outputs.valid_type = Data

        dynamic_class_name = persistence.get_object_loader().identify_object(calc_class)
        class_name = '{}_{}'.format(cls.__name__, dynamic_class_name)

        # Dynamically create the type for this Process
        return type(
            class_name, (cls,),
            {
                plumpy.Process.define.__name__: classmethod(define),
                '_calc_class': calc_class
            }
        )

    @classmethod
    def get_state_classes(cls):
        # Overwrite the waiting state
        states_map = super(JobProcess, cls).get_state_classes()
        states_map[processes.ProcessState.WAITING] = Waiting
        return states_map

    # region Process overrides

    @override
    def on_killed(self):
        super(JobProcess, self).on_killed()
        self.calc._set_state(calc_states.FAILED)

    @override
    def update_outputs(self):
        # DO NOT REMOVE:
        # Don't do anything, this is taken care of by the job calculation node itself
        pass

    @override
    def get_or_create_db_record(self):
        return self._calc_class()

    @property
    def process_class(self):
        """
        Return the class that represents this Process, for the JobProcess this is JobCalculation class it wraps.

        For a standard Process or sub class of Process, this is the class itself. However, for legacy reasons,
        the Process class is a wrapper around another class. This function returns that original class, i.e. the
        class that really represents what was being executed.
        """
        return self._calc_class

    @override
    def _setup_db_inputs(self):
        """
        Create the links that connect the inputs to the calculation node that represents this Process

        For a JobProcess, the inputs also need to be mapped onto the `use_` and `set_` methods of the
        legacy JobCalculation class. If a code is defined in the inputs and no computer has been set
        yet for the calculation node, the computer configured for the code is used to set on the node.
        """
        for name, input_value in self.get_provenance_inputs_iterator():

            port = self.spec().inputs[name]

            if input_value is None or getattr(port, 'non_db', False):
                continue

            # Call the 'set' attribute methods for the contents of the 'option' namespace
            if name == self.OPTIONS_INPUT_LABEL:
                for option_name, option_value in input_value.items():
                    getattr(self.calc, 'set_{}'.format(option_name))(option_value)
                continue

            # Call the 'use' methods to set up the data-calc links
            if isinstance(port, PortNamespace):
                additional = self._calc_class._use_methods[name]['additional_parameter']

                for k, v in input_value.items():
                    try:
                        getattr(self.calc, 'use_{}'.format(name))(v, **{additional: k})
                    except AttributeError:
                        raise AttributeError(
                            "You have provided for an input the key '{}' but"
                            "the JobCalculation has no such use_{} method".format(name, name))

            else:
                getattr(self.calc, 'use_{}'.format(name))(input_value)

        # Get the computer from the code if necessary
        if self.calc.get_computer() is None and 'code' in self.inputs:
            code = self.inputs['code']
            if not code.is_local():
                self.calc.set_computer(code.get_remote_computer())

    # endregion

    @override
    def run(self):
        """
        Run the calculation, we put it in the TOSUBMIT state and then wait for it
        to be copied over, submitted, retrieved, etc.
        """
        from aiida.orm import Code, load_node
        from aiida.common.folders import SandboxFolder
        from aiida.common.exceptions import InputValidationError

        # Note that the caching mechanism relies on this as it will always enter the run method, even when finished
        if self.calc.get_state() == calc_states.FINISHED:
            return 0

        state_current = self.calc.get_state()
        state_pending = calc_states.TOSUBMIT

        if is_progressive_state_change(state_current, state_pending):
            self.calc._set_state(state_pending)
        else:
            logger.warning('ignored invalid proposed state change: {} to {}'.format(state_current, state_pending))

        with SandboxFolder() as folder:
            computer = self.calc.get_computer()
            calc_info, script_filename = self.calc._presubmit(folder, use_unstored_links=False)
            input_codes = [load_node(_.code_uuid, sub_classes=(Code,)) for _ in calc_info.codes_info]

            for code in input_codes:
                if not code.can_run_on(computer):
                    raise InputValidationError(
                        'The selected code {} for calculation {} cannot run on computer {}'.format(
                            code.pk, self.calc.pk, computer.name))

            # After this call, no modifications to the folder should be done
            self.calc._store_raw_input_folder(folder.abspath)

        # Launch the upload operation
        return plumpy.Wait(msg='Waiting to upload', data=(UPLOAD_COMMAND, calc_info, script_filename))

    def retrieved(self, retrieved_temporary_folder=None):
        """
        Parse a retrieved job calculation.  This is called once it's finished waiting
        for the calculation to be finished and the data has been retrieved.
        """
        try:
            exit_code = execmanager.parse_results(self.calc, retrieved_temporary_folder)
        except Exception:
            try:
                self.calc._set_state(calc_states.PARSINGFAILED)
            except exceptions.ModificationNotAllowed:
                pass
            raise
        finally:
            # Delete the temporary folder
            try:
                shutil.rmtree(retrieved_temporary_folder)
            except OSError as exception:
                if exception.errno != 2:
                    raise

        # Finally link up the outputs and we're done
        for label, node in self.calc.get_outputs_dict().items():
            self.out(label, node)

        return exit_code


class ContinueJobCalculation(JobProcess):

    @classmethod
    def define(cls, spec):
        super(ContinueJobCalculation, cls).define(spec)
        spec.input('_calc', valid_type=JobCalculation, required=True, non_db=False)

    def run(self):
        state = self.calc.get_state()

        if state == calc_states.NEW:
            return super(ContinueJobCalculation, self).run()

        if state in [calc_states.TOSUBMIT, calc_states.SUBMITTING]:
            return plumpy.Wait(msg='Waiting to submit', data=SUBMIT_COMMAND)

        elif state in calc_states.WITHSCHEDULER:
            return plumpy.Wait(msg='Waiting for scheduler', data=UPDATE_COMMAND)

        elif state in [calc_states.COMPUTED, calc_states.RETRIEVING]:
            return plumpy.Wait(msg='Waiting to retrieve', data=RETRIEVE_COMMAND)

        elif state == calc_states.PARSING:
            return self.retrieved(True)

            # Otherwise nothing to do...

    def get_or_create_db_record(self):
        return self.inputs._calc

    @override
    def _setup_db_record(self):
        self._calc_class = self.inputs._calc.__class__
        super(ContinueJobCalculation, self)._setup_db_record()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""The computer backend abstract classes"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import abc
import logging
import os
import six

from aiida import transport, scheduler
from aiida.common import exceptions
from .backend import Collection, CollectionEntry

__all__ = ['Computer', 'ComputerCollection']


@six.add_metaclass(abc.ABCMeta)
class ComputerCollection(Collection):
    """The collection of Computer entries."""

    @abc.abstractmethod
    def create(self,
               name,
               hostname,
               description='',
               transport_type='',
               scheduler_type='',
               workdir=None,
               enabled_state=True):
        # pylint: disable=too-many-arguments
        """
        Create new a computer

        :return: the newly created computer
        :rtype: :class:`aiida.orm.Computer`
        """
        pass

    def get(self, id=None, name=None, uuid=None):  # pylint: disable=redefined-builtin, invalid-name
        """
        Get a computer from one of it's unique identifiers

        :param id: the computer's id
        :param name: the name of the computer
        :type name: str
        :param uuid: the uuid of the computer
        :return: the corresponding computer
        :rtype: :class:`aiida.orm.Computer`
        """
        query = self.backend.query_builder()
        filters = {}
        if id is not None:
            filters['id'] = {'==': id}
        if name is not None:
            filters['name'] = {'==': name}
        if uuid is not None:
            filters['uuid'] = {'==': uuid}

        query.append(Computer, filters=filters)
        res = [_[0] for _ in query.all()]
        if not res:
            raise exceptions.NotExistent("No computer with filter '{}' found".format(filters))
        if len(res) > 1:
            raise exceptions.MultipleObjectsError("Multiple computers found with the same id '{}'".format(id))

        return res[0]

    @abc.abstractmethod
    def list_names(self):
        """
        Return a list with all the names of the computers in the DB.
        """
        pass

    @abc.abstractmethod
    def delete(self, id):  # pylint: disable=invalid-name, redefined-builtin
        """ Delete the computer with the given id """
        pass


@six.add_metaclass(abc.ABCMeta)
class Computer(CollectionEntry):
    """
    Base class to map a node in the DB + its permanent repository counterpart.

    Stores attributes starting with an underscore.

    Caches files and attributes before the first save, and saves everything only on store().
    After the call to store(), attributes cannot be changed.

    Only after storing (or upon loading from uuid) metadata can be modified
    and in this case they are directly set on the db.

    In the plugin, also set the _plugin_type_string, to be set in the DB in the 'type' field.
    """
    # pylint: disable=too-many-public-methods

    _logger = logging.getLogger(__name__)

    PROPERTY_MINIMUM_SCHEDULER_POLL_INTERVAL = 'minimum_scheduler_poll_interval'  # pylint: disable=invalid-name
    PROPERTY_MINIMUM_SCHEDULER_POLL_INTERVAL__DEFAULT = 10.  # pylint: disable=invalid-name

    @staticmethod
    def get_schema():
        """
        Every node property contains:
            - display_name: display name of the property
            - help text: short help text of the property
            - is_foreign_key: is the property foreign key to other type of the node
            - type: type of the property. e.g. str, dict, int

        :return: get schema of the computer
        """
        return {
            "description": {
                "display_name": "Description",
                "help_text": "short description of the Computer",
                "is_foreign_key": False,
                "type": "str"
            },
            "enabled": {
                "display_name": "Enabled",
                "help_text": "True(False) if the computer is(not) enabled to run jobs",
                "is_foreign_key": False,
                "type": "bool"
            },
            "hostname": {
                "display_name": "Host",
                "help_text": "Name of the host",
                "is_foreign_key": False,
                "type": "str"
            },
            "id": {
                "display_name": "Id",
                "help_text": "Id of the object",
                "is_foreign_key": False,
                "type": "int"
            },
            "name": {
                "display_name": "Name",
                "help_text": "Name of the object",
                "is_foreign_key": False,
                "type": "str"
            },
            "scheduler_type": {
                "display_name": "Scheduler",
                "help_text": "Scheduler type",
                "is_foreign_key": False,
                "type": "str",
                "valid_choices": {
                    "direct": {
                        "doc": "Support for the direct execution bypassing schedulers."
                    },
                    "pbsbaseclasses.PbsBaseClass": {
                        "doc": "Base class with support for the PBSPro scheduler"
                    },
                    "pbspro": {
                        "doc": "Subclass to support the PBSPro scheduler"
                    },
                    "sge": {
                        "doc":
                        "Support for the Sun Grid Engine scheduler and its variants/forks (Son of Grid Engine, "
                        "Oracle Grid Engine, ...)"
                    },
                    "slurm": {
                        "doc": "Support for the SLURM scheduler (http://slurm.schedmd.com/)."
                    },
                    "torque": {
                        "doc": "Subclass to support the Torque scheduler.."
                    }
                }
            },
            "transport_params": {
                "display_name": "",
                "help_text": "Transport Parameters",
                "is_foreign_key": False,
                "type": "str"
            },
            "transport_type": {
                "display_name": "Transport type",
                "help_text": "Transport Type",
                "is_foreign_key": False,
                "type": "str",
                "valid_choices": {
                    "local": {
                        "doc":
                        "Support copy and command execution on the same host on which AiiDA is running via direct file "
                        "copy and execution commands."
                    },
                    "ssh": {
                        "doc":
                        "Support connection, command execution and data transfer to remote computers via SSH+SFTP."
                    }
                }
            },
            "uuid": {
                "display_name": "Unique ID",
                "help_text": "Universally Unique Identifier",
                "is_foreign_key": False,
                "type": "unicode"
            }
        }

    def pk(self):
        """
        Return the principal key in the DB.
        """
        return self.id

    @abc.abstractproperty
    def uuid(self):
        """
        Return the UUID in the DB.
        """
        pass

    @abc.abstractproperty
    def id(self):  # pylint: disable=invalid-name
        """
        Return the ID in the DB.
        """
        pass

    @abc.abstractmethod
    def set(self, **kwargs):
        pass

    @abc.abstractproperty
    def full_text_info(self):
        """
        Return a (multiline) string with a human-readable detailed information on this computer.

        :rypte: str
        """
        pass

    @abc.abstractproperty
    def is_stored(self):
        """
        Is the computer stored?

        :return: True if stored, False otherwise
        :rtype: bool
        """

    @property
    def logger(self):
        return self._logger

    @classmethod
    def _name_validator(cls, name):
        """
        Validates the name.
        """
        if not name.strip():
            raise exceptions.ValidationError("No name specified")

    @classmethod
    def _hostname_validator(cls, hostname):
        """
        Validates the hostname.
        """
        if not hostname.strip():
            raise exceptions.ValidationError("No hostname specified")

    @classmethod
    def _enabled_state_validator(cls, enabled_state):
        """
        Validates the hostname.
        """
        if not isinstance(enabled_state, bool):
            raise exceptions.ValidationError("Invalid value '{}' for the enabled state, must "
                                             "be a boolean".format(str(enabled_state)))

    @classmethod
    def _description_validator(cls, description):
        """
        Validates the description.
        """
        # The description is always valid
        pass

    @classmethod
    def _transport_type_validator(cls, transport_type):
        """
        Validates the transport string.
        """
        if transport_type not in transport.Transport.get_valid_transports():
            raise exceptions.ValidationError("The specified transport is not a valid one")

    @classmethod
    def _scheduler_type_validator(cls, scheduler_type):
        """
        Validates the transport string.
        """
        if scheduler_type not in scheduler.Scheduler.get_valid_schedulers():
            raise exceptions.ValidationError("The specified scheduler is not a valid one")

    @classmethod
    def _prepend_text_validator(cls, prepend_text):
        """
        Validates the prepend text string.
        """
        # no validation done
        pass

    @classmethod
    def _append_text_validator(cls, append_text):
        """
        Validates the append text string.
        """
        # no validation done
        pass

    @classmethod
    def _workdir_validator(cls, workdir):
        """
        Validates the transport string.
        """
        if not workdir.strip():
            raise exceptions.ValidationError("No workdir specified")

        try:
            convertedwd = workdir.format(username="test")
        except KeyError as exc:
            raise exceptions.ValidationError("In workdir there is an unknown replacement field {}".format(exc.args[0]))
        except ValueError as exc:
            raise exceptions.ValidationError("Error in the string: '{}'".format(exc))

        if not os.path.isabs(convertedwd):
            raise exceptions.ValidationError("The workdir must be an absolute path")

    def _mpirun_command_validator(self, mpirun_cmd):
        """
        Validates the mpirun_command variable. MUST be called after properly
        checking for a valid scheduler.
        """
        if not isinstance(mpirun_cmd, (tuple, list)) or not all(isinstance(i, six.string_types) for i in mpirun_cmd):
            raise exceptions.ValidationError("the mpirun_command must be a list of strings")

        try:
            job_resource_keys = self.get_scheduler().job_resource_class.get_valid_keys()
        except exceptions.MissingPluginError:
            raise exceptions.ValidationError("Unable to load the scheduler for this computer")

        subst = {i: 'value' for i in job_resource_keys}
        subst['tot_num_mpiprocs'] = 'value'

        try:
            for arg in mpirun_cmd:
                arg.format(**subst)
        except KeyError as exc:
            raise exceptions.ValidationError("In workdir there is an unknown replacement field {}".format(exc.args[0]))
        except ValueError as exc:
            raise exceptions.ValidationError("Error in the string: '{}'".format(exc))

    def validate(self):
        """
        Check if the attributes and files retrieved from the DB are valid.
        Raise a ValidationError if something is wrong.

        Must be able to work even before storing: therefore, use the get_attr and similar methods
        that automatically read either from the DB or from the internal attribute cache.

        For the base class, this is always valid. Subclasses will reimplement this.
        In the subclass, always call the super().validate() method first!
        """
        if not self.get_name().strip():
            raise exceptions.ValidationError("No name specified")

        self._hostname_validator(self.get_hostname())
        self._description_validator(self.get_description())
        self._enabled_state_validator(self.is_enabled())
        self._transport_type_validator(self.get_transport_type())
        self._scheduler_type_validator(self.get_scheduler_type())
        self._workdir_validator(self.get_workdir())

        try:
            mpirun_cmd = self.get_mpirun_command()
        except exceptions.DbContentError:
            raise exceptions.ValidationError("Error in the DB content of the transport_params")

        # To be called AFTER the validation of the scheduler
        self._mpirun_command_validator(mpirun_cmd)

    @classmethod
    def _default_mpiprocs_per_machine_validator(cls, def_cpus_per_machine):
        """
        Validates the default number of CPUs per machine (node)
        """
        if def_cpus_per_machine is None:
            return

        if not isinstance(def_cpus_per_machine, six.integer_types) or def_cpus_per_machine <= 0:
            raise exceptions.ValidationError("Invalid value for default_mpiprocs_per_machine, "
                                             "must be a positive integer, or an empty "
                                             "string if you do not want to provide a "
                                             "default value.")

    @abc.abstractmethod
    def copy(self):
        """
        Return a copy of the current object to work with, not stored yet.
        """
        pass

    @abc.abstractmethod
    def store(self):
        """
        Store the computer in the DB.

        Differently from Nodes, a computer can be re-stored if its properties
        are to be changed (e.g. a new mpirun command, etc.)
        """
        pass

    @abc.abstractproperty
    def name(self):
        pass

    @property
    def label(self):
        """
        The computer label
        """
        return self.name

    @label.setter
    def label(self, value):
        """
        Set the computer label (i.e., name)
        """
        self.set_name(value)

    @abc.abstractproperty
    def description(self):
        pass

    @abc.abstractproperty
    def hostname(self):
        pass

    @abc.abstractmethod
    def _get_metadata(self):
        pass

    @abc.abstractmethod
    def _set_metadata(self, metadata_dict):
        """
        Set the metadata.

        .. note: You still need to call the .store() method to actually save
           data to the database! (The store method can be called multiple
           times, differently from AiiDA Node objects).
        """
        pass

    def _del_property(self, name, raise_exception=True):
        """
        Delete a property from this computer

        :param name: the name of the property
        :param raise_exception: if True raise if the property does not exist, otherwise return None
        """
        olddata = self._get_metadata()
        try:
            del olddata[name]
        except KeyError:
            if raise_exception:
                raise AttributeError("'{}' property not found".format(name))
            else:
                # Do not reset the metadata, it is not necessary
                return
        self._set_metadata(olddata)

    def _set_property(self, name, value):
        """
        Set a property on this computer

        :param name: the property name
        :param value: the new value
        """
        olddata = self._get_metadata()
        olddata[name] = value
        self._set_metadata(olddata)

    def _get_property(self, name, *args):
        """
        Get a property of this computer

        :param name: the property name
        :param args: additional arguments
        :return: the property value
        """
        if len(args) > 1:
            raise TypeError("_get_property expected at most 2 arguments")
        olddata = self._get_metadata()
        try:
            return olddata[name]
        except KeyError:
            if not args:
                raise AttributeError("'{}' property not found".format(name))
            elif len(args) == 1:
                return args[0]

    def get_prepend_text(self):
        return self._get_property("prepend_text", "")

    def set_prepend_text(self, val):
        self._set_property("prepend_text", six.text_type(val))

    def get_append_text(self):
        return self._get_property("append_text", "")

    def set_append_text(self, val):
        self._set_property("append_text", six.text_type(val))

    def get_mpirun_command(self):
        """
        Return the mpirun command. Must be a list of strings, that will be
        then joined with spaces when submitting.

        I also provide a sensible default that may be ok in many cases.
        """
        return self._get_property("mpirun_command", ["mpirun", "-np", "{tot_num_mpiprocs}"])

    def set_mpirun_command(self, val):
        """
        Set the mpirun command. It must be a list of strings (you can use
        string.split() if you have a single, space-separated string).
        """
        if not isinstance(val, (tuple, list)) or not all(isinstance(i, six.string_types) for i in val):
            raise TypeError("the mpirun_command must be a list of strings")
        self._set_property("mpirun_command", val)

    def get_default_mpiprocs_per_machine(self):
        """
        Return the default number of CPUs per machine (node) for this computer,
        or None if it was not set.
        """
        return self._get_property("default_mpiprocs_per_machine", None)

    def set_default_mpiprocs_per_machine(self, def_cpus_per_machine):
        """
        Set the default number of CPUs per machine (node) for this computer.
        Accepts None if you do not want to set this value.
        """
        if def_cpus_per_machine is None:
            self._del_property("default_mpiprocs_per_machine", raise_exception=False)
        else:
            if not isinstance(def_cpus_per_machine, six.integer_types):
                raise TypeError("def_cpus_per_machine must be an integer (or None)")
        self._set_property("default_mpiprocs_per_machine", def_cpus_per_machine)

    def get_minimum_job_poll_interval(self):
        """
        Get the minimum interval between subsequent requests to update the list
        of jobs currently running on this computer.

        :return: The minimum interval (in seconds)
        :rtype: float
        """
        return self._get_property(self.PROPERTY_MINIMUM_SCHEDULER_POLL_INTERVAL,
                                  self.PROPERTY_MINIMUM_SCHEDULER_POLL_INTERVAL__DEFAULT)

    def set_minimum_job_poll_interval(self, interval):
        """
        Set the minimum interval between subsequent requests to update the list
        of jobs currently running on this computer.

        :param interval: The minimum interval in seconds
        :type interval: float
        """
        self._set_property(self.PROPERTY_MINIMUM_SCHEDULER_POLL_INTERVAL, interval)

    @abc.abstractmethod
    def get_transport_params(self):
        pass

    @abc.abstractmethod
    def set_transport_params(self, val):
        pass

    def get_transport(self, user=None):
        """
        Return a Tranport class, configured with all correct parameters.
        The Transport is closed (meaning that if you want to run any operation with
        it, you have to open it first (i.e., e.g. for a SSH tranport, you have
        to open a connection). To do this you can call ``transport.open()``, or simply
        run within a ``with`` statement::

           transport = Computer.get_transport()
           with transport:
               print(transport.whoami())

        :param user: if None, try to obtain a transport for the default user.
            Otherwise, pass a valid User.

        :return: a (closed) Transport, already configured with the connection
            parameters to the supercomputer, as configured with ``verdi computer configure``
            for the user specified as a parameter ``user``.
        """
        from aiida.orm.backend import construct_backend
        backend = construct_backend()
        if user is None:
            authinfo = backend.authinfos.get(self, backend.users.get_automatic_user())
        else:
            authinfo = backend.authinfos.get(self, user)

        return authinfo.get_transport()

    @abc.abstractmethod
    def get_workdir(self):
        """
        Get the working directory for this computer
        :return: The currently configured working directory
        :rtype: str
        """
        pass

    @abc.abstractmethod
    def get_shebang(self):
        pass

    @abc.abstractmethod
    def set_workdir(self, val):
        pass

    def set_shebang(self, val):
        """
        :param str val: A valid shebang line
        """
        if not isinstance(val, six.string_types):
            raise ValueError("{} is invalid. Input has to be a string".format(val))
        if not val.startswith('#!'):
            raise ValueError("{} is invalid. A shebang line has to start with #!".format(val))
        metadata = self._get_metadata()
        metadata['shebang'] = val
        self._set_metadata(metadata)

    @abc.abstractmethod
    def get_name(self):
        pass

    @abc.abstractmethod
    def set_name(self, val):
        pass

    def get_hostname(self):
        """
        Get this computer hostname
        :rtype: str
        """
        pass

    @abc.abstractmethod
    def set_hostname(self, val):
        """
        Set the hostname of this computer
        :param val: The new hostname
        :type val: str
        """
        pass

    @abc.abstractmethod
    def get_description(self):
        """
        Get the description for this computer

        :return: the description
        :rtype: str
        """
        pass

    @abc.abstractmethod
    def set_description(self, val):
        """
        Set the description for this computer

        :param val: the new description
        :type val: str
        """
        pass

    @abc.abstractmethod
    def get_calculations_on_computer(self):
        """
        Get the calculations corresponding to this computer

        :return: a list of calculations on this computer
        """
        pass

    @abc.abstractmethod
    def is_enabled(self):
        """
        Is the computer enabled

        :return: True if enabled, False otherwise
        :rtype: bool
        """
        pass

    def get_authinfo(self, user):
        """
        Return the aiida.orm.authinfo.AuthInfo instance for the
        given user on this computer, if the computer
        is configured for the given user.

        :param user: a User instance.
        :return: a AuthInfo instance
        :raise NotExistent: if the computer is not configured for the given
            user.
        """
        return self.backend.authinfos.get(computer=self, user=user)

    def is_user_configured(self, user):
        """
        Check if the user is configured on this computer

        :param user: the user to check
        :return: True if enabled, False otherwise
        """
        try:
            self.get_authinfo(user)
            return True
        except exceptions.NotExistent:
            return False

    def is_user_enabled(self, user):
        """
        Check if a user is enabled to run on this computer

        :param user: the user to check
        :return: True if enabled, False otherwise
        """
        try:
            authinfo = self.get_authinfo(user)
            return authinfo.enabled
        except exceptions.NotExistent:
            # Return False if the user is not configured (in a sense,
            # it is disabled for that user)
            return False

    @abc.abstractmethod
    def set_enabled_state(self, enabled):
        """
        Set the enable state for this computer

        :param enabled: True if enabled, False otherwise
        """

    @abc.abstractmethod
    def get_scheduler_type(self):
        """
        Get the scheduler type for this computer

        :return: the scheduler ype
        :rtype: str
        """
        pass

    @abc.abstractmethod
    def set_scheduler_type(self, val):
        """
        Set the scheduler type for this computer

        :param val: the new scheduler type
        :type val: str
        """
        pass

    @abc.abstractmethod
    def get_transport_type(self):
        """
        Get the transport type for this computer

        :return: the transport type
        :rtype: str
        """

    @abc.abstractmethod
    def set_transport_type(self, val):
        """
        Set the transport type for this computer

        :param val: the new transport type
        :type val: str
        """
        pass

    def get_transport_class(self):
        """
        Get the transport class for this computer.  Can be used to instantiate a transport instance.

        :return: the transport class
        """
        try:
            # I return the class, not an instance
            return transport.TransportFactory(self.get_transport_type())
        except exceptions.MissingPluginError as exc:
            raise exceptions.ConfigurationError('No transport found for {} [type {}], message: {}'.format(
                self.name, self.get_transport_type(), exc))

    def get_scheduler(self):
        """
        Get the scheduler instance for this computer

        :return: the scheduler
        """
        try:
            scheduler_class = scheduler.SchedulerFactory(self.get_scheduler_type())
            # I call the init without any parameter
            return scheduler_class()
        except exceptions.MissingPluginError as exc:
            raise exceptions.ConfigurationError('No scheduler found for {} [type {}], message: {}'.format(
                self.name, self.get_scheduler_type(), exc))

    def configure(self, user=None, **kwargs):
        """
        Configure a computer for a user with valid auth params passed via kwargs

        :param user: the user to configure the computer for
        :kwargs: the configuration keywords with corresponding values
        :return: the authinfo object for the configured user
        :rtype: :class:`aiida.orm.AuthInfo`
        """

        transport_cls = self.get_transport_class()
        backend = self.backend
        user = user or backend.users.get_automatic_user()

        try:
            authinfo = self.get_authinfo(user)
        except exceptions.NotExistent:
            authinfo = backend.authinfos.create(self, user)

        auth_params = authinfo.get_auth_params()
        valid_keys = set(transport_cls.get_valid_auth_params())

        if not set(kwargs.keys()).issubset(valid_keys):
            invalid_keys = [key for key in kwargs if key not in valid_keys]
            raise ValueError('{transport}: recieved invalid authentication parameter(s) "{invalid}"'.format(
                transport=transport_cls, invalid=invalid_keys))

        if valid_keys:
            auth_params.update(kwargs)
            authinfo.set_auth_params(auth_params)
            authinfo.store()

        return authinfo

    def get_configuration(self, user=None):
        """
        Get the configuration of computer for the given user as a dictionary

        :param user: the user to to get the configuration for.  Uses default user if `None`
        :type user: :class:`aiida.orm.User`
        """

        backend = self.backend
        user = user or backend.users.get_automatic_user()

        config = {}
        try:
            authinfo = backend.authinfos.get(self, user)
            config = authinfo.get_auth_params()
        except exceptions.NotExistent:
            pass

        return config

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, str(self))

    def __str__(self):
        if self.is_enabled():
            return "{} ({}), pk: {}".format(self.name, self.hostname, self.pk)

        return "{} ({}) [DISABLED], pk: {}".format(self.name, self.hostname, self.pk)
